"""Interactive tests for pyscal, supposed to be looked at by a human.

Run pytest with --plot option to make these run. Add -s also to get
instructions for what to check."""

import random

import matplotlib
import numpy as np
import pytest
from matplotlib import pyplot

from pyscal import GasOil, GasWater, WaterOil, utils
from pyscal.factory import create_scal_recommendation

from .test_scalrecommendation import BASE_SAMPLE_LET, HIGH_SAMPLE_LET, LOW_SAMPLE_LET


@pytest.mark.plot
def test_interpolation_art():
    """This code was used to create the Pyscal logo (50 repeats and 30 interpolants)"""
    repeats = 15
    interpolants = 30
    curvetype = "corey"

    cmap = pyplot.get_cmap("viridis")
    _, mpl_ax = pyplot.subplots()
    for _ in range(repeats):
        swl = random.uniform(0, 0.1)
        swcr = swl + random.uniform(0, 0.1)
        sorw = random.uniform(0, 0.2)
        wo_low = WaterOil(swl=swl, swcr=swcr, sorw=sorw)
        wo_high = WaterOil(swl=swl + 0.1, swcr=swcr + 0.1, sorw=sorw + 0.1)
        if curvetype == "corey":
            wo_low.add_corey_water(
                nw=random.uniform(1, 3), krwend=random.uniform(0.5, 1)
            )
            wo_high.add_corey_water(
                nw=random.uniform(1, 3), krwend=random.uniform(0.5, 1)
            )
            wo_low.add_corey_oil(
                now=random.uniform(1, 3), kroend=random.uniform(0.5, 1)
            )
            wo_high.add_corey_oil(
                now=random.uniform(1, 3), kroend=random.uniform(0.5, 1)
            )
        elif curvetype == "let":
            wo_low.add_LET_water(
                l=random.uniform(1, 3),
                e=random.uniform(1, 3),
                t=random.uniform(1, 3),
                krwend=random.uniform(0.5, 1),
            )
            wo_high.add_LET_water(
                l=random.uniform(1, 3),
                e=random.uniform(1, 3),
                t=random.uniform(1, 3),
                krwend=random.uniform(0.5, 1),
            )
            wo_low.add_LET_oil(
                l=random.uniform(1, 3),
                e=random.uniform(1, 3),
                t=random.uniform(1, 3),
                kroend=random.uniform(0.5, 1),
            )
            wo_high.add_LET_oil(
                l=random.uniform(1, 3),
                e=random.uniform(1, 3),
                t=random.uniform(1, 3),
                kroend=random.uniform(0.5, 1),
            )
        else:
            print("ERROR, wrong curvetype")
        color = cmap(random.random())
        for tparam in np.arange(0, 1, 1.0 / interpolants):
            wo_ip = utils.interpolation.interpolate_wo(wo_low, wo_high, tparam)
            wo_ip.plotkrwkrow(mpl_ax, color=color, alpha=0.3)
    pyplot.title("Pyscal art")
    pyplot.show()


def _pyplot_show_with_user_message():
    print(
        "Ensure green curves are within red and blue, "
        "and check behaviour around endpoints"
    )
    print("[Close windows to continue tests]")
    pyplot.show()


@pytest.mark.plot
@pytest.mark.parametrize("", [(), (), ()])  # 3 repeated runs
def test_interpolate_wo():
    """Discrete test scenarios for wateroil interpolation"""
    swl_l = random.uniform(0, 0.1)
    swcr_l = swl_l + random.uniform(0, 0.1)
    sorw_l = random.uniform(0, 0.2)
    socr_l = sorw_l + random.uniform(0, 0.3)
    swl_h = random.uniform(0, 0.1)
    swcr_h = swl_h + random.uniform(0, 0.1)
    sorw_h = random.uniform(0, 0.2)
    socr_h = sorw_h + random.uniform(0, 0.3)
    wo_low = WaterOil(swl=swl_l, swcr=swcr_l, sorw=sorw_l, socr=socr_l, h=0.001)
    wo_high = WaterOil(swl=swl_h, swcr=swcr_h, sorw=sorw_h, socr=socr_h, h=0.001)
    wo_low.add_corey_water(nw=random.uniform(1, 3), krwend=random.uniform(0.5, 1))
    wo_high.add_corey_water(nw=random.uniform(1, 3), krwend=random.uniform(0.5, 1))
    wo_low.add_corey_oil(now=random.uniform(1, 3), kroend=random.uniform(0.5, 1))
    wo_high.add_corey_oil(now=random.uniform(1, 3), kroend=random.uniform(0.5, 1))
    wo_low.add_simple_J(a=random.uniform(0.1, 2), b=random.uniform(-2, -1))
    wo_high.add_simple_J(a=random.uniform(0.1, 2), b=random.uniform(-2, 1))
    print(
        " ** Low curve WaterOil (red):\n"
        + wo_low.swcomment
        + wo_low.krwcomment
        + wo_low.krowcomment
        + wo_low.pccomment
    )
    print(
        " ** High curve WaterOil (blue):\n"
        + wo_high.swcomment
        + wo_high.krwcomment
        + wo_high.krowcomment
        + wo_high.pccomment
    )

    _, mpl_ax = pyplot.subplots()
    wo_low.plotkrwkrow(mpl_ax, color="red")
    wo_high.plotkrwkrow(mpl_ax, color="blue")
    for tparam in np.arange(0, 1, 0.1):
        wo_ip = utils.interpolation.interpolate_wo(wo_low, wo_high, tparam, h=0.001)
        wo_ip.plotkrwkrow(mpl_ax, color="green")
    mpl_ax.set_title("WaterOil, random Corey, linear y-scale")
    _pyplot_show_with_user_message()

    # Plot again with log yscale:
    _, mpl_ax = pyplot.subplots()
    wo_low.plotkrwkrow(mpl_ax, color="red")
    wo_high.plotkrwkrow(mpl_ax, color="blue")
    for tparam in np.arange(0, 1, 0.1):
        wo_ip = utils.interpolation.interpolate_wo(wo_low, wo_high, tparam, h=0.001)
        wo_ip.plotkrwkrow(mpl_ax, color="green", logyscale=True)
    mpl_ax.set_title("WaterOil, random Corey, log y-scale")
    _pyplot_show_with_user_message()

    # Capillary pressure
    _, mpl_ax = pyplot.subplots()
    wo_low.plotpc(mpl_ax, color="red", logyscale=True)
    wo_high.plotpc(mpl_ax, color="blue", logyscale=True)
    for tparam in np.arange(0, 1, 0.1):
        wo_ip = utils.interpolation.interpolate_wo(wo_low, wo_high, tparam, h=0.001)
        wo_ip.plotpc(mpl_ax, color="green", logyscale=True)
    mpl_ax.set_title("WaterOil, capillary pressure")
    _pyplot_show_with_user_message()


@pytest.mark.plot
@pytest.mark.parametrize("", [(), (), ()])  # 3 repeated runs
def test_interpolate_go():
    """Interactive tests for gasoil"""
    swl_l = random.uniform(0, 0.1)
    sgcr_l = random.uniform(0, 0.1)
    swl_h = random.uniform(0, 0.1)
    sgcr_h = random.uniform(0, 0.1)
    sorg_l = random.uniform(0, 0.2)
    sorg_h = random.uniform(0, 0.2)
    if bool(random.getrandbits(1)):
        # Interpolation is not possible if only
        # one of the curves has nonzero sgro
        sgro_l = sgcr_l
        sgro_h = sgcr_h
    else:
        sgro_l = 0
        sgro_h = 0
    krgend_l = random.uniform(0.5, 1)
    krgend_h = random.uniform(0.5, 1)
    kromax_l = random.uniform(0.5, 1)
    kromax_h = random.uniform(0.5, 1)
    kroend_l = min(random.uniform(0.5, 1), kromax_l)
    kroend_h = min(random.uniform(0.5, 1), kromax_h)
    krgendanchor_l = "sorg" if random.uniform(0, 1) > 0.5 else ""
    krgendanchor_h = "sorg" if random.uniform(0, 1) > 0.5 else ""
    go_low = GasOil(
        swl=swl_l,
        sgcr=sgcr_l,
        sorg=sorg_l,
        sgro=sgro_l,
        krgendanchor=krgendanchor_l,
        h=0.001,
    )
    go_high = GasOil(
        swl=swl_h,
        sgcr=sgcr_h,
        sorg=sorg_h,
        sgro=sgro_h,
        krgendanchor=krgendanchor_h,
        h=0.001,
    )
    go_low.add_corey_gas(ng=random.uniform(1, 3), krgend=krgend_l)
    go_high.add_corey_gas(ng=random.uniform(1, 3), krgend=krgend_h)

    go_low.add_corey_oil(nog=random.uniform(1, 3), kroend=kroend_l, kromax=kromax_l)
    go_high.add_corey_oil(nog=random.uniform(1, 3), kroend=kroend_h, kromax=kromax_h)
    print(
        " ** Low curve GasOil (red):\n"
        + go_low.sgcomment
        + go_low.krgcomment
        + go_low.krogcomment
    )
    print(
        " ** High curve GasOil (blue):\n"
        + go_high.sgcomment
        + go_high.krgcomment
        + go_high.krogcomment
    )

    _, mpl_ax = pyplot.subplots()
    go_low.plotkrgkrog(mpl_ax, color="red")
    go_high.plotkrgkrog(mpl_ax, color="blue")

    for tparam in np.arange(0, 1, 0.1):
        go_ip = utils.interpolation.interpolate_go(go_low, go_high, tparam)
        go_ip.plotkrgkrog(mpl_ax, color="green")
    mpl_ax.set_title("GasOil, random Corey, linear y-scale")
    _pyplot_show_with_user_message()

    _, mpl_ax = pyplot.subplots()
    go_low.plotkrgkrog(mpl_ax, color="red")
    go_high.plotkrgkrog(mpl_ax, color="blue")
    # Plot again with log yscale:
    for tparam in np.arange(0, 1, 0.1):
        go_ip = utils.interpolation.interpolate_go(go_low, go_high, tparam)
        go_ip.plotkrgkrog(mpl_ax, color="green", logyscale=True)
    mpl_ax.set_title("GasOil, random Corey, log y-scale")
    _pyplot_show_with_user_message()


@pytest.mark.plot
@pytest.mark.parametrize("", [(), (), ()])  # 3 repeated runs
def test_interpolate_gw():
    """Discrete test scenarios for gaswater interpolation"""
    swl_l = random.uniform(0, 0.1)
    swcr_l = swl_l + random.uniform(0, 0.1)
    sgrw_l = random.uniform(0, 0.2)
    sgcr_l = random.uniform(0, 0.3)
    swl_h = random.uniform(0, 0.1)
    swcr_h = swl_h + random.uniform(0, 0.1)
    sgrw_h = random.uniform(0, 0.2)
    sgcr_h = random.uniform(0, 0.3)
    gw_low = GasWater(swl=swl_l, swcr=swcr_l, sgrw=sgrw_l, sgcr=sgcr_l, h=0.001)
    gw_high = GasWater(swl=swl_h, swcr=swcr_h, sgrw=sgrw_h, sgcr=sgcr_h, h=0.001)
    gw_low.add_corey_water(nw=random.uniform(1, 3), krwend=random.uniform(0.5, 1))
    gw_high.add_corey_water(nw=random.uniform(1, 3), krwend=random.uniform(0.5, 1))
    gw_low.add_corey_gas(ng=random.uniform(1, 3), krgend=random.uniform(0.5, 1))
    gw_high.add_corey_gas(ng=random.uniform(1, 3), krgend=random.uniform(0.5, 1))
    print(
        " ** Low curve GasWater (red):\n"
        + gw_low.swcomment
        + gw_low.krwcomment
        + gw_low.krgcomment
    )
    print(
        " ** High curve GasWater (blue):\n"
        + gw_high.swcomment
        + gw_high.krwcomment
        + gw_high.krgcomment
    )

    _, mpl_ax = pyplot.subplots()
    gw_low.plotkrwkrg(mpl_ax, color="red")
    gw_high.plotkrwkrg(mpl_ax, color="blue")
    for tparam in np.arange(0, 1, 0.1):
        gw_wo_ip = utils.interpolation.interpolate_wo(
            gw_low.wateroil, gw_high.wateroil, tparam, h=0.001
        )
        gw_go_ip = utils.interpolation.interpolate_go(
            gw_low.gasoil, gw_high.gasoil, tparam, h=0.001
        )
        gw_ip = GasWater()
        gw_ip.gasoil = gw_go_ip
        gw_ip.wateroil = gw_wo_ip
        gw_ip.plotkrwkrg(mpl_ax, color="green")
    mpl_ax.set_title("GasWater, random Corey, linear y-scale")
    _pyplot_show_with_user_message()

    # Plot again with log yscale:
    _, mpl_ax = pyplot.subplots()
    gw_low.plotkrwkrg(mpl_ax, color="red")
    gw_high.plotkrwkrg(mpl_ax, color="blue")
    for tparam in np.arange(0, 1, 0.1):
        gw_wo_ip = utils.interpolation.interpolate_wo(
            gw_low.wateroil, gw_high.wateroil, tparam, h=0.001
        )
        gw_go_ip = utils.interpolation.interpolate_go(
            gw_low.gasoil, gw_high.gasoil, tparam, h=0.001
        )
        gw_ip = GasWater()
        gw_ip.gasoil = gw_go_ip
        gw_ip.wateroil = gw_wo_ip
        gw_ip.plotkrwkrg(mpl_ax, color="green", logyscale=True)
    mpl_ax.set_title("GasWater, random Corey, log y-scale")
    _pyplot_show_with_user_message()


@pytest.mark.plot
def test_SCAL_interpolation():
    """Demonstration of interpolation between LET curves, 2x2 subplot"""
    matplotlib.style.use("ggplot")

    rec = create_scal_recommendation(
        {"low": LOW_SAMPLE_LET, "base": BASE_SAMPLE_LET, "high": HIGH_SAMPLE_LET},
        "FOO",
        h=0.001,
    )
    _, ((ax1, ax2), (ax3, ax4)) = pyplot.subplots(2, 2)

    # Choosing logarithmic spaced interpolation parameters
    # is not the same as interpolating in log(kr)-space
    # check the effect by setting
    #  for t in -2 + np.logspace(1e-5,1e-1,15):
    # and
    #  for t in -1 + np.logspace(1e-5,1e-1,15)
    # in the loops below. Curves get clustered to the bottom
    # in both linear and log(kr) spaces, but there
    # still might be some other distribution for the interpolants
    # that yields something that spans nicely both the linear and the
    # logarithmic kr space (?)

    for tparam in np.arange(-1, 0, 0.2):
        interp = rec.interpolate(tparam, h=0.001)
        interp.wateroil.plotkrwkrow(ax1, "r")
        interp.wateroil.plotkrwkrow(ax2, "r")

    for tparam in np.arange(0, 1, 0.2):
        interp = rec.interpolate(tparam, h=0.001)
        interp.wateroil.plotkrwkrow(ax1, "g")
        interp.wateroil.plotkrwkrow(ax2, "g")

    rec.low.wateroil.plotkrwkrow(ax1, linewidth=2, linestyle=":")
    rec.base.wateroil.plotkrwkrow(ax1, linewidth=2)
    rec.high.wateroil.plotkrwkrow(ax1, linewidth=2, linestyle="--")
    rec.low.wateroil.plotkrwkrow(ax2, linewidth=2, linestyle=":")
    rec.base.wateroil.plotkrwkrow(ax2, linewidth=2)
    rec.high.wateroil.plotkrwkrow(ax2, linewidth=2, linestyle="--")
    ax2.set_yscale("log")
    ax2.set_ylim([1e-10, 1])
    ax1.set_xlabel("")
    ax2.set_xlabel("")
    ax1.set_title("Water-oil, low, base, high and interpolants", fontsize=10)
    ax2.set_title("Water-oil, low, base, high and interpolants", fontsize=10)

    for tparam in np.arange(-1, 0, 0.2):
        interp = rec.interpolate(tparam, h=0.001)
        interp.gasoil.plotkrgkrog(ax3, "r")
        interp.gasoil.plotkrgkrog(ax4, "r")

    for tparam in np.arange(0, 1, 0.2):
        interp = rec.interpolate(tparam, h=0.001)
        interp.gasoil.plotkrgkrog(ax3, "g")
        interp.gasoil.plotkrgkrog(ax4, "g")

    rec.low.gasoil.plotkrgkrog(ax3, linewidth=2, linestyle=":")
    rec.base.gasoil.plotkrgkrog(ax3, linewidth=2)
    rec.high.gasoil.plotkrgkrog(ax3, linewidth=2, linestyle="--")
    rec.low.gasoil.plotkrgkrog(ax4, linewidth=2, linestyle=":")
    rec.base.gasoil.plotkrgkrog(ax4, linewidth=2)
    rec.high.gasoil.plotkrgkrog(ax4, linewidth=2, linestyle="--")
    ax3.set_title("Gas-oil, low, base, high and interpolants", fontsize=10)
    ax4.set_title("Gas-oil, low, base, high and interpolants", fontsize=10)
    ax4.set_yscale("log")
    ax4.set_ylim([1e-05, 1])
    ax3.set_xlabel("")
    ax4.set_xlabel("")
    pyplot.subplots_adjust(hspace=0.3)
    print("--  Check:")
    print("--   * Red curves are between dotted and solid blue line")
    print("--   * Green curves are between solid blue and dashed")
    print("[Close windows to continue tests]")
    pyplot.show()


@pytest.mark.plot
def test_letspan():
    """Demonstration of how random LET (random individually and
    uncorrelated in L, eand T) curves span out the relperm-space
    between low and high LET curves

    If the low and high LET curves do not cross, the random
    curves are all between the low and high curves.
    """
    matplotlib.style.use("ggplot")

    let_w = {
        "l": [2, 4],
        "e": [1, 2],
        "t": [2, 1],  # first value should be larger than first to avoid crossing
        "krwend": [0.9, 0.5],
        "sorw": [0.05, 0.1],
    }

    # Parameter test set from SCAL group, first is pessimistic, second
    # is optimistic
    let_w = {
        "l": [2.323, 4.436],
        "e": [2, 8],
        "t": [1.329, 0.766],  # first value should be larger
        # than first to avoid crossing
        "krwend": [0.9, 0.6],
        "sorw": [0.02, 0.137],
    }

    # LET oil:
    let_o = {
        "l": [4.944, 2.537],
        "e": [5, 2],
        "t": [0.68, 1.549],  # first value should be larger
        # than first to avoid crossing
        "kroend": [1, 1],
        "sorw": [0.02, 0.137],
    }

    # We need sorted versions for the random function
    slimw = {x: sorted(value) for x, value in let_w.items()}
    slimo = {x: sorted(value) for x, value in let_o.items()}

    _, mpl_ax = pyplot.subplots()
    for _ in range(100):
        swof = WaterOil(
            h=0.01, swl=0.16, sorw=random.uniform(slimw["sorw"][0], slimw["sorw"][1])
        )
        swof.add_LET_water(
            l=random.uniform(slimw["l"][0], slimw["l"][1]),
            e=random.uniform(slimw["e"][0], slimw["e"][1]),
            t=random.uniform(slimw["t"][0], slimw["t"][1]),
            krwend=random.uniform(slimw["krwend"][0], slimw["krwend"][1]),
        )
        swof.add_LET_oil(
            l=random.uniform(slimo["l"][0], slimo["l"][1]),
            e=random.uniform(slimo["e"][0], slimo["e"][1]),
            t=random.uniform(slimo["t"][0], slimo["t"][1]),
        )
        swof.plotkrwkrow(mpl_ax=mpl_ax, alpha=0.1)
    # Boundary lines
    swof = WaterOil(h=0.01, sorw=let_w["sorw"][0], swl=0.16)
    swof.add_LET_water(
        l=let_w["l"][0], e=let_w["e"][0], t=let_w["t"][0], krwend=let_w["krwend"][0]
    )
    swof.add_LET_oil(l=let_o["l"][0], e=let_o["e"][0], t=let_o["t"][0])
    swof.plotkrwkrow(mpl_ax=mpl_ax, color="red", label="Low")
    swof = WaterOil(h=0.01, sorw=let_w["sorw"][1], swl=0.16)
    swof.add_LET_water(
        l=let_w["l"][1], e=let_w["e"][1], t=let_w["t"][1], krwend=let_w["krwend"][1]
    )
    swof.add_LET_oil(l=let_o["l"][1], e=let_o["e"][1], t=let_o["t"][1])
    swof.plotkrwkrow(mpl_ax=mpl_ax, color="red", label="High")

    pyplot.subplots_adjust(top=0.8)
    pyplot.title(
        "Blue LET curves go outside red boundaries, \n"
        "therefore SCAL-interpolation "
        "must be done\n on curves, not in LET-space.",
        fontsize=12,
    )
    print("[Close windows to continue tests]")
    pyplot.show()



"""Utility functions for creating strings from pyscal"""

import logging
from typing import Dict, Optional

import pandas as pd

from .monotonicity import MonotonicitySpec, modify_dframe_monotonicity

logger = logging.getLogger(__name__)


def df2str(
    dframe: pd.DataFrame,
    digits: int = 7,
    roundlevel: int = 9,
    header: bool = False,
    monotonicity: Optional[Dict[str, MonotonicitySpec]] = None,
) -> str:
    """
    Make a string representation of a dataframe with
    proper rounding.

    This is used to print the tables in the SWOF/SGOF include files,
    explicit rounding is necessary to avoid monotonicity errors
    from truncation. Examples in test code.

    Capillary pressure must be strictly monotone if nonzero, and if a
    column name is provided, the string representation of that column is
    ensured to be strictly monotone decreasing

    Args:
        dframe: A dataframe to print, all columns
            are included
        digits: Number of digits used in floating point format f.ex ".7f"
            It is not recommended to deviate from the default 7 uncritically
            for pyscal output, other code have to be tuned to ensure
            numerical robustness to the deviation.
        roundlevel: To how many digits should we round prior to print.
            Recommended to be > digits + 1, see test code.
        header: If the dataframe column header should be included
        monotonicity: Column names in dframe are the keys, pointing
            to a specification for monotonicity to be enforced.
    """
    float_format = "%1." + str(digits) + "f"

    if monotonicity is not None:
        dframe = modify_dframe_monotonicity(dframe, monotonicity, digits)

    return dframe.round(roundlevel).to_csv(
        sep=" ", float_format=float_format, header=header, index=False
    )


def comment_formatter(multiline: str, prefix: str = "-- "):
    """Prepends comment characters to every line in input

    Args:
        multiline: String that can contain newlines
        prefix: Comment characters to prepend every line with
            Default is the Eclipse comment syntax '-- '

    Returns:
        string with newlines preserved, and where each line
        starts with the given prefix. Always ends with a newline.
    """
    if multiline is None or not multiline.strip():
        # Ensure we indicate that there is placeholder for something.
        return "-- \n"
    return (
        "\n".join([prefix + line.strip() for line in multiline.splitlines()]).strip()
        + "\n"
    )



"""Module for plotting relative permeability and capillary pressure curves.

Potential improvements:
    Plotting low/base/high curves before interpolation

    Plotting curve sets on the same plot, e.g. curves from different SATNUMs or
    low/base/high curves from a SCAL recommendation

    Option to plot GasWater curves with Sg axis (instead of Sw; the Sg column
    is already present in the dataframe)

"""

from pathlib import Path

import matplotlib.pyplot as plt
import pandas as pd

from pyscal import GasOil, GasWater, PyscalList, WaterOil, WaterOilGas, getLogger_pyscal

logger = getLogger_pyscal(__name__)

# Data for configuring plot based on pyscal model type
PLOT_CONFIG_OPTIONS = {
    "WaterOil": {
        "axis": "SW",
        "kra_name": "KRW",
        "krb_name": "KROW",
        "kra_colour": "blue",
        "krb_colour": "green",
        "pc_name": "PCOW",
        "xlabel": "Sw",
        "ylabel": "krw, krow",
        "curves": "krw_krow",
    },
    "GasOil": {
        "axis": "SG",
        "kra_name": "KRG",
        "krb_name": "KROG",
        "kra_colour": "red",
        "krb_colour": "green",
        "xlabel": "Sg",
        "ylabel": "krg, krog",
        "pc_name": "PCOG",
        "curves": "krg_krog",
    },
    "GasWater": {
        "axis": "SW",
        "kra_name": "KRW",
        "krb_name": "KRG",
        "kra_colour": "blue",
        "krb_colour": "red",
        "pc_name": "PCGW",
        "xlabel": "Sw",
        "ylabel": "krg, krw",
        "curves": "krg_krw",
    },
}


def format_gaswater_table(model: GasWater) -> pd.DataFrame:
    """

    Format the tables held by the GasWater object (through the GasOil and
    WaterOil objects).

    The GasWater object is a bit trickier to work with. Like WaterOilGas,
    GasWater takes from both WaterOil and GasOil objects, but unlike
    WaterOilGas, it is a two-phase model. It is based on one curve from GasOil,
    krg, and one curve from WaterOil, krw. This is a different format to the
    other models, where the relperm curves to be plotted in the same figure are
    found in the same table and are accessed easily using the "table" instance
    variable, e.g. WaterOil.table. To plot the correct curves for the GasWater
    model, additional formatting is required. That is handled by this function.

    Process: sort both the GasOil and WaterOil tables by increasing SW and then
    merge on index. Can't join on "SW" due to potential floating-point number
    errors.

    Args:
        model (GasWater): GasWater model

    Returns:
        pd.DataFrame: saturation table with Sw, Sg, krg from the GasOil
        instance, and krw and Pc from the WaterOil instance
    """

    gasoil = model.gasoil.table[["SL", "KRG"]].copy()
    gasoil = gasoil.rename(columns={"SL": "SW"})
    gasoil["SG"] = 1 - gasoil["SW"]
    wateroil = model.wateroil.table[["SW", "KRW", "PC"]].copy()

    # Sort by SW to be able to join on index
    gasoil = gasoil.sort_values("SW", ascending=True)
    wateroil = wateroil.sort_values("SW", ascending=True)
    gasoil = gasoil.reset_index(drop=True)
    wateroil = wateroil.reset_index(drop=True)

    # Check if SW in the two models differs using an epsilon of 1E-6 If any
    # absolute differences are greater than this threshold, an assertion error
    # will be raised.
    assert (abs(gasoil["SW"] - wateroil["SW"]) < 1e-6).all(), (
        "SW for the GasOil model does not match SW for the WaterOil model"
    )

    # Merge dataframes and format
    gaswater = gasoil.merge(wateroil, left_index=True, right_index=True)
    gaswater = gaswater.rename(columns={"SW_x": "SW"})
    return gaswater.drop("SW_y", axis=1)


def get_satnum_from_tag(string: str) -> int:
    """
    Get SATNUM from the model tag. Used in the naming of figures.

    Args:
        string (str): String from the model .tag instance variable

    Returns:
        int: SATNUM number
    """
    return int(string.split("SATNUM")[1].strip())


def get_plot_config_options(curve_type: str, **kwargs) -> dict:
    """
    Get config data from plot config dictionary based on the curve (model) type.

    Args:
        curve_type (str): Name of the curve type. Allowed types are given in
        the PLOT_CONFIG_OPTIONS dictionary

    Returns:
        dict: Config parameters for the chosen model type
    """

    config = PLOT_CONFIG_OPTIONS[curve_type].copy()

    # If semilog plot, add suffix to the name of the saved relperm figure
    suffix = "_semilog" if kwargs["semilog"] else ""

    config["suffix"] = suffix

    return config


def format_relperm_plot(fig: plt.Figure, **kwargs) -> plt.Figure:
    """
    Formatting options for individual relative permeability plots.

    Args:
        fig (plt.Figure): Relative permeability figure to be formatted

    Returns:
        plt.Figure: Formatted relative permeability figure
    """

    ax = fig.gca()

    # Set axis lables
    ax.set_xlabel(kwargs["xlabel"])
    ax.set_ylabel(kwargs["ylabel"])

    # Set axis limits
    ax.set_xlim((0.0, 1.0))

    # Log y-axis fro semilog plot
    # Some cases may have kr > 1 depending on base/absolute
    # permeability used to calculate kr
    if kwargs["semilog"]:
        ax.set_yscale("log")
        ax.set_ylim((1e-6, 1.0))
    else:
        ax.set_ylim((0.0, 1.0))

    # Add legend
    plt.legend()
    ax.legend(
        loc="upper center",
        bbox_to_anchor=(0.5, -0.12),
        ncol=5,
    )

    return fig


def format_cap_pressure_plot(
    fig: plt.Figure, neg_pc: bool = False, **kwargs
) -> plt.Figure:
    """
    Formatting options for individual capillary pressure plots.

    Args:
        fig (plt.Figure): Capillary pressure figure to be formatted
        neg_pc (bool, optional): Negative Pc flag. True if the Pc curve has a
        negative part. Defaults to False.

    Returns:
        plt.Figure: Formatted capillary pressure figure
    """

    ax = fig.gca()

    # Set axis lables
    ax.set_xlabel(kwargs["xlabel"])
    ax.set_ylabel(kwargs["pc_name"].lower().capitalize())

    # Set axis limits
    ax.set_xlim((0.0, 1.0))

    # Set lower y-axis limit to 0 if Pc >= 0
    if not neg_pc:
        ax.set_ylim(bottom=0.0)

    return fig


def plot_pc(table: pd.DataFrame, satnum: int, **kwargs) -> plt.Figure:
    """
    Plot capillary pressure curves.

    Called if the Pc plot is requested, regardless of if Pc is non-zero.

    Args:
        table (pd.DataFrame): Saturation table with Pc curves to be plotted
        satnum (int): SATNUM number

    Returns:
        plt.Figure: Pc figure with a single Pc curve for a given model and
        SATNUM
    """

    fig = plt.figure(1, figsize=(5, 5), dpi=300)
    plt.title(f"SATNUM {satnum}")
    plt.plot(table[kwargs["axis"]], table["PC"])

    # Flag for negative Pc
    neg_pc = table["PC"].min() < 0
    fig = format_cap_pressure_plot(fig, neg_pc, **kwargs)

    # Log warning if Pc plot is requested but is zero or practically zero.
    # There is no checking of units of Pc in pyscal, but this should work as
    # intended for Pc in bar, Pa, MPa, atm and psi, i.e. the most common units
    # of pressure. In any case, the figures will still be made.
    if not (abs(table["PC"]) > 1e-6).any():
        logger.warning("Pc plots were requested, but Pc is zero.")

    return fig


def plot_relperm(
    table: pd.DataFrame, satnum: int, config: dict, **kwargs
) -> plt.Figure:
    """

    Function for plotting one relperm curve set for each SATNUM.

    Takes kwargs from the plotter() function, which in turn come from the
    pyscal CLI, and are passed on to the plot formatting functions.

    Args:
        table (pd.DataFrame): Saturation table with curves to be plotted
        satnum (int): SATNUM number
        config (dict): Plot config

    Returns:
        plt.Figure: Relative permeability figure with two relative permeability
        curves for a given model and SATNUM
    """

    # Plotting relative permeability curves
    fig = plt.figure(1, figsize=(5, 5), dpi=300)

    plt.title(f"SATNUM {satnum}")

    # Plot first relperm curve
    plt.plot(
        table[config["axis"]],
        table[config["kra_name"]],
        label=config["kra_name"].lower(),
        color=config["kra_colour"],
    )

    # Plot second relperm curve
    plt.plot(
        table[config["axis"]],
        table[config["krb_name"]],
        label=config["krb_name"].lower(),
        color=config["krb_colour"],
    )

    return format_relperm_plot(fig, **kwargs, **config)


def save_figure(
    fig: plt.Figure,
    satnum: int,
    config: dict,
    plot_type: str,
    outdir: str,
) -> None:
    """

    Save the provided figure.

    Args:
        fig (plt.Figure): Figure to be saved
        satnum (int): SATNUM number
        config (dict): Plot config
        plot_type (str): Figure type. Allowed types are 'relperm' and 'pc'
        outdir (str): Directory where the figure will be saved
    """

    # Get curve name
    if plot_type == "relperm":
        curve_names = config["curves"]
        suffix = config["suffix"]
    elif plot_type == "pc":
        curve_names = config["pc_name"].lower()
        suffix = ""
    else:
        raise ValueError(f"'type' given ({plot_type}) must be one of 'relperm' or 'pc'")

    fname = f"{curve_names}_SATNUM_{satnum}{suffix}".replace(" ", "")
    fout = Path(outdir).joinpath(fname)

    fig.savefig(
        fout,
        bbox_inches="tight",
    )

    print(f"Figure saved to {fout}.png")

    # Clear figure so that it is empty for the next SATNUM's plot
    fig.clear()


def wog_plotter(model: WaterOilGas, **kwargs) -> None:
    """

    Plot a WaterOilGas (WaterOil and GasOil) model.


    For a WaterOilGas instance, the WaterOil and GasOil instances can be
    accessed, then the "table" instance variable.

    Args:
        model (WaterOilGas): WaterOilGas instance
    """

    outdir = kwargs["outdir"]

    # the wateroil and gasoil instance variables are optional for the
    # WaterOilGas class. If statements used to check if they are provided
    if model.wateroil:
        config_wo = get_plot_config_options("WaterOil", **kwargs)
        satnum_wo = get_satnum_from_tag(model.wateroil.tag)

        fig_wo = plot_relperm(
            model.wateroil.table,
            satnum_wo,
            config_wo,
            **kwargs,
        )

        save_figure(fig_wo, satnum_wo, config_wo, "relperm", outdir)

        if kwargs["pc"]:
            fig_pcwo = plot_pc(
                model.wateroil.table,
                get_satnum_from_tag(model.wateroil.tag),
                **config_wo,
            )

            save_figure(fig_pcwo, satnum_wo, config_wo, "pc", outdir)

    if model.gasoil:
        config_go = get_plot_config_options("GasOil", **kwargs)
        satnum_go = get_satnum_from_tag(model.gasoil.tag)

        fig_go = plot_relperm(
            model.gasoil.table,
            satnum_go,
            config_go,
            **kwargs,
        )

        save_figure(fig_go, satnum_go, config_go, "relperm", outdir)


def wo_plotter(model: WaterOil, **kwargs) -> None:
    """

    Plot a WaterOil model.

    For a WaterOil instance, the saturation table can be accessed using the
    "table" instance variable.

    Args:
        model (WaterOil): WaterOil instance
    """
    config = get_plot_config_options("WaterOil", **kwargs)
    satnum = get_satnum_from_tag(model.tag)

    outdir = kwargs["outdir"]

    fig = plot_relperm(
        model.table,
        satnum,
        config,
        **kwargs,
    )

    save_figure(fig, satnum, config, "relperm", outdir)

    if kwargs["pc"]:
        fig_pc = plot_pc(model.table, get_satnum_from_tag(model.tag), **config)

        save_figure(fig_pc, satnum, config, "pc", outdir)


def go_plotter(model: GasOil, **kwargs) -> None:
    """

    Plot a GasOil model.

    For a GasOil instance, the saturation table can be accessed using the
    "table" instance variable.

    Args:
        model (GasOil): GasOil instance
    """

    config = get_plot_config_options("GasOil", **kwargs)
    satnum = get_satnum_from_tag(model.tag)

    outdir = kwargs["outdir"]

    fig = plot_relperm(
        model.table,
        satnum,
        config,
        **kwargs,
    )

    save_figure(fig, satnum, config, "relperm", outdir)

    # Note that there are no supporting functions for adding Pc to GasOil
    # instances. This can only be done by modifying the "table" instance
    # variable for a GasOil object
    if kwargs["pc"]:
        fig_pc = plot_pc(model.table, get_satnum_from_tag(model.tag), **config)

        save_figure(fig_pc, satnum, config, "pc", outdir)


def gw_plotter(model: GasWater, **kwargs) -> None:
    """

    For GasWater, the format is different, and an additional formatting step is
    required. Use the formatted table as an argument to the plotter function,
    instead of the "table" instance variable

    Args:
        model (GasWater): GasWater instance
    """

    table = format_gaswater_table(model)
    config = get_plot_config_options("GasWater", **kwargs)
    satnum = get_satnum_from_tag(model.tag)
    outdir = kwargs["outdir"]

    fig = plot_relperm(
        table,
        satnum,
        config,
        **kwargs,
    )

    save_figure(fig, satnum, config, "relperm", outdir)

    if kwargs["pc"]:
        fig_pc = plot_pc(table, get_satnum_from_tag(model.tag), **config)

        save_figure(fig_pc, satnum, config, "pc", outdir)


def plotter(
    models: PyscalList, pc: bool = False, semilog: bool = False, outdir: str = "./"
) -> None:
    """

    Runner function for creating plots.

    Iterate over PyscalList and plot curves based on type of pyscal objects
    encountered.

    PyscalList is a list of WaterOilGas, WaterOil, GasOil or GasWater objects.

    For WaterOil and GasOil, the saturation table can be accessed using the
    "table" instance variable.

    For WaterOilGas, the WaterOil and GasOil instances can be accessed, then
    the "table" instance variable.

    For GasWater, the format is different, and an additional formatting step is
    required.

    Args:
        models (PyscalList): List of models
        pc (bool, optional): Plot Pc flag. Defaults to False.
        semilog (bool, optional): Plot relperm with log y-axis. Defaults to
        False.

    """

    # kwargs to be passed on to other functions
    kwargs = {"pc": pc, "semilog": semilog, "outdir": outdir}

    for model in models.pyscal_list:
        if isinstance(model, WaterOilGas):
            wog_plotter(model, **kwargs)
        elif isinstance(model, WaterOil):
            wo_plotter(model, **kwargs)
        elif isinstance(model, GasOil):
            go_plotter(model, **kwargs)
        elif isinstance(model, GasWater):
            gw_plotter(model, **kwargs)
        else:
            raise TypeError(
                f"Model type received was {type(model)} but\
         must be one of: {WaterOil, WaterOilGas, GasOil, GasWater}"
            )



"""Test module for GasOil objects"""

import io
import sys

import hypothesis.strategies as st
import matplotlib
import matplotlib.pyplot
import numpy as np
import pandas as pd
import pytest
from hypothesis import given

from pyscal import GasOil
from pyscal.constants import MAX_EXPONENT_KR, SWINTEGERS
from pyscal.utils.relperm import truncate_zeroness
from pyscal.utils.testing import (
    check_linear_sections,
    check_table,
    float_df_checker,
    sat_table_str_ok,
)


def test_gasoil_init():
    """Test features in the constructor"""
    gasoil = GasOil()
    assert isinstance(gasoil, GasOil)
    assert gasoil.swirr == 0.0
    assert gasoil.swl == 0.0
    assert gasoil.krgendanchor == ""  # Because sorg is zero

    gasoil = GasOil(swl=0.1)
    assert gasoil.swirr == 0.0
    assert gasoil.swl == 0.1

    gasoil = GasOil(swirr=0.1)
    assert gasoil.swirr == 0.1
    assert gasoil.swl == 0.1  # This one is zero by default, but will follow swirr.
    assert gasoil.sorg == 0.0
    assert gasoil.sgcr == 0.0

    gasoil = GasOil(tag="foobar")
    assert gasoil.tag == "foobar"

    # This will print a warning, but will be the same as ""
    gasoil = GasOil(krgendanchor="bogus")
    assert isinstance(gasoil, GasOil)
    assert gasoil.krgendanchor == ""

    gasoil = GasOil(krgendanchor=None)
    assert isinstance(gasoil, GasOil)
    assert gasoil.krgendanchor == ""

    # Test with h=1
    gasoil = GasOil(h=1)
    gasoil.add_corey_gas()
    gasoil.add_corey_oil()
    assert np.isclose(gasoil.crosspoint(), 0.5)
    assert len(gasoil.table) == 2

    gasoil = GasOil(swl=0.1, h=1)
    gasoil.add_corey_gas()
    gasoil.add_corey_oil()
    assert len(gasoil.table) == 2
    assert np.isclose(gasoil.crosspoint(), 0.45)
    assert np.isclose(gasoil.table["SG"].min(), 0)
    assert np.isclose(gasoil.table["SG"].max(), 0.9)

    # Test too small h:
    gasoil = GasOil(swl=0.1, h=0.00000000000000000001)
    # (a warning is printed that h is truncated)
    assert gasoil.h == 10 / SWINTEGERS


def test_conserve_sgcr(mocker):
    """sgcr in the table can be hard to conserve"""
    low_sgcrvalue = 1 / SWINTEGERS / 10
    # Verify the default behaviour of truncation:
    assert truncate_zeroness(low_sgcrvalue) == 0
    mocker.patch(
        # Remove the effect of zeroness truncation in order to provoke
        # the safeguarding measures for sgcr.
        "pyscal.gasoil.truncate_zeroness",
        return_value=low_sgcrvalue,
    )
    gasoil = GasOil(swl=0, sgcr=low_sgcrvalue, h=0.2)
    assert gasoil.sgcr == low_sgcrvalue  # To check that we have not truncated it away
    assert gasoil.table["SG"][0] == 0
    assert gasoil.table["SG"][1] == low_sgcrvalue

    gasoil.add_corey_gas()
    gasoil.add_corey_oil()
    assert gasoil.selfcheck()
    assert "SGOF" in gasoil.SGOF()


def test_errors():
    """Test some error situations for the constructor"""
    with pytest.raises(ValueError, match="No saturation range left"):
        GasOil(swl=0.3, sorg=0.8)
    with pytest.raises(ValueError, match="No saturation range left"):
        GasOil(swl=0.3, sgro=0.88, sgcr=0.88)
    with pytest.raises(ValueError, match="No saturation range left"):
        GasOil(swl=0.3, sgcr=0.88)

    # This only raises error if krgendanchor is non-default:
    GasOil(swl=0.3, sgcr=0.4, sorg=0.4, krgendanchor="")
    with pytest.raises(ValueError, match="No saturation range left"):
        GasOil(swl=0.3, sgcr=0.4, sorg=0.4)

    with pytest.raises(ValueError, match="No saturation range left"):
        GasOil(swl=0.3, sgcr=0.8, sorg=0.4, krgendanchor="")

    with pytest.raises(ValueError, match="No saturation range left"):
        GasOil(swl=0.3, sgcr=0.88, krgendanchor="")
    with pytest.raises(ValueError, match="sgro must be zero or equal to sgcr"):
        GasOil(swl=0.3, sgcr=0.1, sgro=0.2)

    with pytest.raises(ValueError, match="sgro must be zero or equal to sgcr"):
        GasOil(swl=0.4, sorg=0.4, sgro=0.4)


@pytest.mark.skipif(sys.platform == "win32", reason="Plotting not stable on windows")
def test_plotting(mocker):
    """Test that plotting code pass through (nothing displayed)"""
    mocker.patch("matplotlib.pyplot.show", return_value=None)
    gasoil = GasOil(swl=0.1, h=0.1)
    gasoil.add_corey_gas()
    gasoil.add_corey_oil()
    gasoil.plotkrgkrog(mpl_ax=matplotlib.pyplot.subplots()[1])
    gasoil.plotkrgkrog(mpl_ax=None)
    gasoil.plotkrgkrog(logyscale=True, mpl_ax=None)


@given(st.text())
def test_gasoil_tag(tag):
    """Test tagging of GasOil objects,
    that we are not able to produce something that
    can crash Eclipse"""
    gasoil = GasOil(h=0.5, tag=tag)
    gasoil.add_corey_gas()
    gasoil.add_corey_oil()
    sat_table_str_ok(gasoil.SGOF())
    sat_table_str_ok(gasoil.SGFN())


@given(
    st.floats(min_value=0, max_value=0.15),  # swl
    st.floats(min_value=0, max_value=0.3),  # sgcr
    st.floats(min_value=0, max_value=0.05),  # sorg
    st.floats(min_value=0.0001, max_value=0.2),  # h
    st.text(),
)
def test_gasoil_normalization(swl, sgcr, sorg, h, tag):
    """Check that normalization (sgn and son) is correct
    for all possible saturation endpoints"""
    gasoil = GasOil(
        swirr=0.0, swl=swl, sgcr=sgcr, sorg=sorg, h=h, krgendanchor="sorg", tag=tag
    )
    assert not gasoil.table.empty
    assert not gasoil.table.isna().to_numpy().any()

    # Check that son is 1 at sg=0
    assert float_df_checker(gasoil.table, "SG", 0, "SON", 1)

    # Check that son is 0 at sorg with this krgendanchor.
    # It is important to use the endpoints from the returned gasoil
    # object, as they can be modified in case they are too close to zero.
    assert float_df_checker(gasoil.table, "SG", 1 - gasoil.sorg - gasoil.swl, "SON", 0)

    # Check that sgn is 0 at sgcr
    assert float_df_checker(gasoil.table, "SG", gasoil.sgcr, "SGN", 0)

    # Check that sgn is 1 at sorg
    assert float_df_checker(gasoil.table, "SG", 1 - gasoil.sorg - gasoil.swl, "SGN", 1)

    # Redo with different krgendanchor
    gasoil = GasOil(
        swirr=0.0, swl=swl, sgcr=sgcr, sorg=sorg, h=h, krgendanchor="", tag=tag
    )
    assert float_df_checker(gasoil.table, "SG", 1 - gasoil.swl, "SGN", 1)
    assert float_df_checker(gasoil.table, "SG", gasoil.sgcr, "SGN", 0)


@given(
    st.floats(min_value=0, max_value=0.3),  # swl
    st.floats(min_value=0, max_value=0.3),  # sgcr
    st.floats(min_value=0, max_value=0.4),  # sorg (sgn collapses when >0.4)
    st.booleans(),  # sgrononzero
    st.floats(min_value=0.1, max_value=1),  # kroend
    st.floats(min_value=0.1, max_value=1),  # kromax
    st.floats(min_value=0.1, max_value=1),  # krgend
    st.floats(min_value=0.2, max_value=1),  # krgmax
    st.floats(min_value=0.001, max_value=0.5),  # h
    st.booleans(),  # fast mode
)
def test_gasoil_krendmax(
    swl, sgcr, sorg, sgrononzero, kroend, kromax, krgend, krgmax, h, fast
):
    """Test that relperm curves are valid in all numerical corner cases."""
    sgro = sgcr if sgrononzero else 0
    try:
        assert 1 - sorg - swl - sgcr > 1 / SWINTEGERS, "No saturation range left"
        gasoil = GasOil(
            swl=swl, sgcr=sgcr, sorg=sorg, sgro=sgro, h=h, tag="", fast=fast
        )
    except AssertionError:
        # We end here when hypothesis sets up impossible/non-interesting scenarios
        return
    krgend = min(krgend, krgmax)
    kroend = min(kroend, kromax)
    gasoil.add_corey_oil(kroend=kroend, kromax=kromax)
    gasoil.add_corey_gas(krgend=krgend, krgmax=krgmax)
    check_table(gasoil.table)
    check_linear_sections(gasoil)
    assert gasoil.selfcheck()
    sat_table_str_ok(gasoil.SGOF())
    check_endpoints(gasoil, krgend, krgmax, kroend, kromax)
    assert 0 < gasoil.crosspoint() < 1

    # Redo with krgendanchor not defaulted
    gasoil = GasOil(swl=swl, sgcr=sgcr, sorg=sorg, h=h, krgendanchor="", tag="")
    gasoil.add_corey_oil(kroend=kroend)
    gasoil.add_corey_gas(krgend=krgend, krgmax=krgmax)
    check_table(gasoil.table)
    check_linear_sections(gasoil)
    assert gasoil.selfcheck()
    check_endpoints(gasoil, krgend, krgmax, kroend, kromax)
    assert 0 < gasoil.crosspoint() < 1

    # Redo with LET:
    gasoil = GasOil(swl=swl, sgcr=sgcr, sorg=sorg, h=h, tag="")
    gasoil.add_LET_oil(t=1.1, kroend=kroend, kromax=kromax)
    gasoil.add_LET_gas(krgend=krgend, krgmax=krgmax)
    check_table(gasoil.table)
    check_linear_sections(gasoil)
    assert gasoil.selfcheck()
    check_endpoints(gasoil, krgend, krgmax, kroend, kromax)
    assert 0 < gasoil.crosspoint() < 1


def check_endpoints(gasoil, krgend, krgmax, kroend, kromax):
    """Discrete tests that endpoints get numerically correct"""
    swtol = 1 / SWINTEGERS

    # Oil curve, from sg = 0 to sg = 1:
    if gasoil.sgro > 0:
        # Gas-condensate: sgcr = sgro > 0
        assert float_df_checker(gasoil.table, "SG", 0, "KROG", kromax)
        assert float_df_checker(gasoil.table, "SON", 1.0, "KROG", kroend)
        assert np.isclose(gasoil.table["KROG"].max(), kromax)
    else:
        assert float_df_checker(gasoil.table, "SG", 0, "KROG", kroend)
        assert np.isclose(gasoil.table["KROG"].max(), kroend)

    # son=0 @ 1 - sorg - swl or 1 - swl) should be zero:
    assert float_df_checker(gasoil.table, "SON", 0.0, "KROG", 0)
    # sgn=1 @ 1 - swl:
    assert float_df_checker(gasoil.table, "SGN", 1.0, "KROG", 0)
    assert np.isclose(gasoil.table["KROG"].min(), 0.0)

    # Gas curve, from sg=0 to sg=1:
    assert float_df_checker(gasoil.table, "SG", 0.0, "KRG", 0)
    assert float_df_checker(gasoil.table, "SGN", 0.0, "KRG", 0)
    assert float_df_checker(gasoil.table, "SG", gasoil.sgcr, "KRG", 0)

    # If krgendanchor == "sorg" then krgmax is irrelevant.
    if gasoil.sorg > swtol and gasoil.sorg > gasoil.h and gasoil.krgendanchor == "sorg":
        assert float_df_checker(gasoil.table, "SGN", 1.0, "KRG", krgend)
        assert np.isclose(gasoil.table["KRG"].max(), krgmax)
    if gasoil.krgendanchor != "sorg":
        assert np.isclose(gasoil.table["KRG"].max(), krgend)
    assert np.isclose(gasoil.table["KRG"].min(), 0.0)


def test_sgro_vs_sgcr():
    """Test that a sgro that is close enough to sgcr does not cause problems"""
    with pytest.raises(ValueError, match="sgro must be zero or equal to sgcr"):
        GasOil(sgcr=0.1, sgro=0.1 + 1e-7, h=0.01)
    with pytest.raises(ValueError, match="sgro must be zero or equal to sgcr"):
        GasOil(sgcr=0.1, sgro=0.1 - 1e-7, h=0.01)

    # This is close enough to sgcr:
    gasoil = GasOil(sgcr=0.1, sgro=0.1 + 1e-8, h=0.01)
    gasoil.add_corey_oil(nog=2, kroend=0.5, kromax=1)

    # KROG should not be affected at sg=0.1, even if sgro is "larger":
    assert float_df_checker(gasoil.table, "SG", 0.0, "KROG", 1.0)
    assert float_df_checker(gasoil.table, "SG", 0.1, "KROG", 0.5)
    assert gasoil.table["SG"][0] == 0
    assert gasoil.table["SG"][1] == 0.1


def test_gasoil_krgendanchor():
    """Test behaviour of the krgendanchor"""
    gasoil = GasOil(krgendanchor="sorg", sorg=0.2, h=0.1)
    assert gasoil.sorg
    gasoil.add_corey_gas(ng=1)
    gasoil.add_corey_oil(nog=1)

    # kg should be 1.0 at 1 - sorg due to krgendanchor == "sorg":
    assert (
        gasoil.table[np.isclose(gasoil.table["SG"], 1 - gasoil.sorg)]["KRG"].to_numpy()[
            0
        ]
        == 1.0
    )
    assert gasoil.table[np.isclose(gasoil.table["SG"], 1.0)]["KRG"].to_numpy()[0] == 1.0

    gasoil = GasOil(krgendanchor="", sorg=0.2, h=0.1)
    assert gasoil.sorg
    gasoil.add_corey_gas(ng=1)
    gasoil.add_corey_oil(nog=1)

    # kg should be < 1 at 1 - sorg due to krgendanchor being ""
    assert (
        gasoil.table[np.isclose(gasoil.table["SG"], 1 - gasoil.sorg)]["KRG"].to_numpy()[
            0
        ]
        < 1.0
    )
    assert gasoil.table[np.isclose(gasoil.table["SG"], 1.0)]["KRG"].to_numpy()[0] == 1.0
    assert gasoil.selfcheck()
    assert gasoil.crosspoint() > 0

    # Test once more for LET curves:
    gasoil = GasOil(krgendanchor="sorg", sorg=0.2, h=0.1)
    assert gasoil.sorg
    gasoil.add_LET_gas(1, 1, 1.1)
    gasoil.add_LET_oil(1, 1, 1.1)
    check_linear_sections(gasoil)
    assert 0 < gasoil.crosspoint() < 1

    # kg should be 1.0 at 1 - sorg due to krgendanchor == "sorg":
    assert (
        gasoil.table[np.isclose(gasoil.table["SG"], 1 - gasoil.sorg)]["KRG"].to_numpy()[
            0
        ]
        == 1.0
    )
    assert gasoil.table[np.isclose(gasoil.table["SG"], 1.0)]["KRG"].to_numpy()[0] == 1.0

    gasoil = GasOil(krgendanchor="", sorg=0.2, h=0.1)
    assert gasoil.sorg
    gasoil.add_LET_gas(1, 1, 1.1)
    gasoil.add_LET_oil(1, 1, 1.1)
    check_linear_sections(gasoil)
    assert gasoil.selfcheck()

    # kg should be < 1 at 1 - sorg due to krgendanchor being ""
    assert (
        gasoil.table[np.isclose(gasoil.table["SG"], 1 - gasoil.sorg)]["KRG"].to_numpy()[
            0
        ]
        < 1.0
    )
    assert gasoil.table[np.isclose(gasoil.table["SG"], 1.0)]["KRG"].to_numpy()[0] == 1.0


def test_nexus():
    """Test the Nexus export"""
    gasoil = GasOil(h=0.01, swl=0.1, sgcr=0.3, sorg=0.3)
    gasoil.add_corey_oil(nog=10, kroend=0.5)
    gasoil.add_corey_gas(ng=10, krgend=0.5)
    nexus_lines = gasoil.GOTABLE().splitlines()
    non_comments = [
        line for line in nexus_lines if not line.startswith("!") or not line
    ]
    assert non_comments[0] == "GOTABLE"
    assert non_comments[1] == "SG KRG KROG PC"
    df = pd.read_table(
        io.StringIO("\n".join(non_comments[2:])),
        engine="python",
        sep=r"\s+",
        header=None,
    )
    assert (df.to_numpy() <= 1.0).all()
    assert (df.to_numpy() >= 0.0).all()


def test_linearsegments():
    """Made for testing the linear segments during
    the resolution of issue #163"""
    gasoil = GasOil(h=0.01, swl=0.1, sgcr=0.3, sorg=0.3)
    gasoil.add_corey_oil(nog=10, kroend=0.5)
    gasoil.add_corey_gas(ng=10, krgend=0.5)
    check_table(gasoil.table)
    check_linear_sections(gasoil)


def test_kroend():
    """Manual testing of kromax and kroend behaviour"""
    gasoil = GasOil(swirr=0.01, sgcr=0.01, h=0.01, swl=0.1, sorg=0.05)
    gasoil.add_LET_gas()
    gasoil.add_LET_oil(2, 2, 2.1)
    assert gasoil.table["KROG"].max() == 1
    gasoil.add_LET_oil(2, 2, 2.1, kroend=0.5)
    check_linear_sections(gasoil)
    assert gasoil.table["KROG"].max() == 0.5

    assert 0 < gasoil.crosspoint() < 1

    gasoil.add_corey_oil(2)
    assert gasoil.table["KROG"].max() == 1
    gasoil.add_corey_oil(nog=2, kroend=0.5)
    assert gasoil.table["KROG"].max() == 0.5


@given(
    st.floats(min_value=1e-4, max_value=MAX_EXPONENT_KR),  # ng
    st.floats(min_value=1e-4, max_value=MAX_EXPONENT_KR),  # nog
)
def test_gasoil_corey1(ng, nog):
    """Test the Corey formulation for gasoil"""
    gasoil = GasOil()
    try:
        gasoil.add_corey_oil(nog=nog)
        gasoil.add_corey_gas(ng=ng)
    except AssertionError:
        # This happens for "invalid" input
        return

    assert "KROG" in gasoil.table
    assert "KRG" in gasoil.table
    assert isinstance(gasoil.krgcomment, str)
    check_table(gasoil.table)
    sgofstr = gasoil.SGOF()
    assert len(sgofstr) > 100
    sat_table_str_ok(sgofstr)

    gasoil.update_sgcomment_and_sorg()
    check_table(gasoil.table)
    check_linear_sections(gasoil)
    sgofstr = gasoil.SGOF()
    assert len(sgofstr) > 100
    sat_table_str_ok(sgofstr)


def test_comments():
    """Test that the outputters include endpoints in comments"""
    gasoil = GasOil(h=0.3)
    gasoil.add_corey_gas()
    gasoil.add_corey_oil()
    sgfn = gasoil.SGFN()
    assert "--" in sgfn
    assert "pyscal: " in sgfn  # part of version string
    assert "swirr=0" in sgfn
    assert "sgcr=0" in sgfn
    assert "swl=0" in sgfn
    assert "sorg=0" in sgfn
    assert "ng=2" in sgfn
    assert "krgend=1" in sgfn
    assert "Corey" in sgfn
    assert "krg = krog @ sg=0.5" in sgfn
    assert "Zero capillary pressure" in sgfn
    assert "SG" in sgfn
    assert "KRG" in sgfn
    assert "PC" in sgfn

    sgof = gasoil.SGOF()
    assert "--" in sgof
    assert "pyscal: " in sgof  # part of version string
    assert "swirr=0" in sgof
    assert "sgcr=0" in sgof
    assert "swl=0" in sgof
    assert "sorg=0" in sgof
    assert "ng=2" in sgof
    assert "nog=2" in sgof
    assert "krgend=1" in sgof
    assert "Corey" in sgof
    assert "krg = krog @ sg=0.5" in sgof
    assert "Zero capillary pressure" in sgof
    assert "SG" in sgof
    assert "KRG" in sgof
    assert "KROG" in sgof
    assert "PC" in sgof


@given(st.floats(), st.floats(), st.floats(), st.floats(), st.floats())
def test_gasoil_let1(l, e, t, krgend, krgmax):
    """Test the LET formulation, take 1"""
    gasoil = GasOil()
    try:
        gasoil.add_LET_oil(l, e, t, krgend)
        gasoil.add_LET_gas(l, e, t, krgend, krgmax)
    except AssertionError:
        # This happens for negative values f.ex.
        return
    assert "KROG" in gasoil.table
    assert "KRG" in gasoil.table
    assert isinstance(gasoil.krgcomment, str)
    check_table(gasoil.table)
    check_linear_sections(gasoil)
    sgofstr = gasoil.SGOF()
    assert len(sgofstr) > 100
    sat_table_str_ok(sgofstr)


def test_sgfn():
    """Test that we can call SGFN without oil relperm defined"""
    gasoil = GasOil()
    gasoil.add_corey_gas()
    sgfn_str = gasoil.SGFN()
    assert "SGFN" in sgfn_str
    assert len(sgfn_str) > 15


def test_fast():
    """Test the fast option"""
    # First without fast-mode:
    gasoil = GasOil(h=0.1)
    gasoil.add_corey_gas()
    gasoil.add_corey_oil()
    # This crosspoint computation is only present for fast=False:
    assert "-- krg = krog @ sg=0.5" in gasoil.SGOF()

    # Provoke non-strict-monotone krow:
    gasoil.table.loc[0:2, "KROG"] = [1.00, 0.81, 0.81]
    # (this is valid in non-imbibition, but pyscal will correct it for all
    # curves)
    assert "0.1000000 0.0100000 0.8100000 0.0000000" in gasoil.SGOF()
    assert "0.2000000 0.0400000 0.8099999 0.0000000" in gasoil.SGOF()
    #   monotonicity correction:   ^^^^^^

    # Now redo with fast option:
    gasoil = GasOil(h=0.1, fast=True)
    gasoil.add_corey_gas()
    gasoil.add_corey_oil()
    # This crosspoint computation is only present for fast=False:
    assert "-- krg = krog" not in gasoil.SGOF()

    # Provoke non-strict-monotone krow, in fast-mode
    # this slips through:
    gasoil.table.loc[0:2, "KROG"] = [1.00, 0.81, 0.81]
    assert "0.1000000 0.0100000 0.8100000 0.0000000" in gasoil.SGOF()
    assert "0.2000000 0.0400000 0.8100000 0.0000000" in gasoil.SGOF()
    # not corrected:               ^^^^^^

    gasoil.table.loc[0:2, "KRG"] = [0.00, 0.01, 0.01]
    assert "0.1000000 0.0100000" in gasoil.SGFN()
    assert "0.2000000 0.0100000" in gasoil.SGFN()


def test_roundoff():
    """Test robustness to monotonicity issues arising from
    representation errors

    https://docs.python.org/3/tutorial/floatingpoint.html#representation-error

    The dataframe injected in this function has occured in the wild, and
    caused fatal errors in Eclipse100. The error lies in
    pd.dataframe.to_csv(float_format=".7f") which does truncation of floating points
    instead of rounding (intentional). Since we have a strict dependency on
    monotonicity properties for Eclipse to work, the data must be rounded
    before being sent to to_csv(). This is being done in the .SGOF() and SWOF() as
    it is a representation issue, not a numerical issues in the objects themselves.
    """

    gasoil = GasOil()
    # Inject a custom dataframe that has occured in the wild,
    # and given monotonicity issues in GasOil.SGOF().
    gasoil.table = pd.DataFrame(
        columns=["SG", "KRG", "KROG", "PC"],
        data=[
            [0.02, 0, 0.19524045000000001, 0],
            [0.040000000000000001, 0, 0.19524044999999998, 0],
            [0.059999999999999998, 0, 0.19524045000000004, 0],
            [0.080000000000000002, 0, 0.19524045000000001, 0],
            [0.10000000000000001, 0, 0.19524045000000001, 0],
            [0.16, 0, 0.19524045000000001, 0],
            [0.17999999999999999, 0, 0.19524045000000001, 0],
            [0.19999999999999998, 0, 0.19524044999999998, 0],
            [0.22, 0, 0.19524045000000001, 0],
            [1, 1, 0, 0],
        ],
    )
    gasoil.table["SGN"] = gasoil.table["SG"]
    gasoil.table["SON"] = 1 - gasoil.table["SG"]
    # If this value (as string) occurs, then we are victim of floating point truncation
    # in float_format=".7f":
    assert "0.1952404" not in gasoil.SGOF()
    assert "0.1952405" in gasoil.SGOF()
    check_table(gasoil.table)  # This function allows this monotonicity hiccup.


@pytest.mark.parametrize(
    "columnname, errorvalues",
    [
        ("SG", [1, 0]),
        ("SG", [0, 2]),
        ("KRG", [1, 0]),
        ("KRG", [0, 2]),
        ("KRG", [1, 2]),
        ("KROG", [0, 1]),
        ("KROG", [2, 0]),
        ("PC", [np.inf, 0]),
        ("PC", [np.nan, 0]),
        ("PC", [1, 2]),
        ("PC", [0, 1]),
    ],
)
def test_selfcheck(columnname, errorvalues):
    """Test the selfcheck feature of a GasOil object"""
    gasoil = GasOil(h=1)
    gasoil.add_corey_gas()
    gasoil.add_corey_oil()
    assert gasoil.selfcheck()

    # Punch the internal table directly to trigger error:
    gasoil.table[columnname] = errorvalues
    assert not gasoil.selfcheck()
    assert gasoil.SGOF() == ""
    if columnname != "KROG":
        assert gasoil.SGFN() == ""



"""Test module for pyscal.utils"""

import pandas as pd

from pyscal.utils.string import comment_formatter, df2str


def test_df2str():
    """Test handling of roundoff issues when printing dataframes

    See also test_gasoil.py::test_roundoff()
    """
    # Easy cases:
    assert df2str(pd.DataFrame(data=[0.1]), digits=1).strip() == "0.1"
    assert df2str(pd.DataFrame(data=[0.1]), digits=3).strip() == "0.100"
    assert df2str(pd.DataFrame(data=[0.1]), digits=3, roundlevel=3).strip() == "0.100"
    assert df2str(pd.DataFrame(data=[0.1]), digits=3, roundlevel=4).strip() == "0.100"
    assert df2str(pd.DataFrame(data=[0.1]), digits=3, roundlevel=5).strip() == "0.100"
    assert df2str(pd.DataFrame(data=[0.01]), digits=3, roundlevel=2).strip() == "0.010"

    # Here roundlevel will ruin the result:
    assert df2str(pd.DataFrame(data=[0.01]), digits=3, roundlevel=1).strip() == "0.000"

    # Tricky ones:
    # This one should be rounded down:
    assert df2str(pd.DataFrame(data=[0.0034999]), digits=3).strip() == "0.003"
    # But if we are on the 9999 side due to representation error, the
    # number probably represents 0.0035 so it should be rounded up
    assert (
        df2str(pd.DataFrame(data=[0.003499999999998]), digits=3, roundlevel=5).strip()
        == "0.004"
    )
    # If we round to more digits than we have in IEE754, we end up truncating:
    assert (
        df2str(pd.DataFrame(data=[0.003499999999998]), digits=3, roundlevel=20).strip()
        == "0.003"  # "Wrong" due to IEE754 truncation.
    )
    # If we round straight to out output, we are not getting the chance to correct for
    # the representation error:
    assert (
        df2str(pd.DataFrame(data=[0.003499999999998]), digits=3, roundlevel=3).strip()
        == "0.003"  # Wrong
    )
    # So roundlevel > digits
    assert (
        df2str(pd.DataFrame(data=[0.003499999999998]), digits=3, roundlevel=4).strip()
        == "0.004"
    )
    # But digits < roundlevel < 15 works:
    assert (
        df2str(pd.DataFrame(data=[0.003499999999998]), digits=3, roundlevel=14).strip()
        == "0.004"
    )
    assert (
        df2str(pd.DataFrame(data=[0.003499999999998]), digits=3, roundlevel=15).strip()
        == "0.003"  # Wrong
    )

    # Double rounding is a potential issue, as:
    assert round(0.0034445, 5) == 0.00344
    assert round(round(0.0034445, 6), 5) == 0.00345  # Wrong
    # So if pd.to_csv would later round instead of truncate, we could be victim
    # of this, having roundlevel >  digits + 1 would avoid that:
    assert round(round(0.0034445, 7), 5) == 0.00344
    # (this is the rationale for roundlevel > digits + 1)


def test_comment_formatter():
    """Test the comment formatter

    This is also tested through hypothesis.text()
    in test_wateroil and test_gasoil, there is it tested
    through the use of tag formatting
    """
    assert comment_formatter(None) == "-- \n"
    assert comment_formatter("\n") == "-- \n"
    assert comment_formatter("\r\n") == "-- \n"
    assert comment_formatter("\r") == "-- \n"
    assert comment_formatter("foo") == "-- foo\n"
    assert comment_formatter("foo", prefix="gaa") == "gaafoo\n"
    assert comment_formatter("foo\nbar") == "-- foo\n-- bar\n"



"""Test module for relperm interpolation support code"""

import hypothesis.strategies as st
import numpy as np
import pandas as pd
import pytest
from hypothesis import given
from matplotlib import pyplot as plt

from pyscal import GasOil, WaterOil
from pyscal.constants import EPSILON as epsilon
from pyscal.utils.interpolation import (
    interpolate_go,
    interpolate_wo,
    normalize_nonlinpart_go,
    normalize_nonlinpart_wo,
    normalize_pc,
)
from pyscal.utils.testing import (
    check_table,
    float_df_checker,
    sat_table_str_ok,
    slow_hypothesis,
)


@slow_hypothesis
@given(
    st.floats(min_value=0, max_value=0.1),  # swl
    st.floats(min_value=0, max_value=0.0),  # dswcr
    st.floats(min_value=0, max_value=0.1),  # dswlhigh
    st.floats(min_value=0, max_value=0.3),  # sorw
    st.floats(min_value=0, max_value=0.2),  # dsocrsorw
    st.floats(min_value=0.1, max_value=5),  # nw1
    st.floats(min_value=0.1, max_value=1),  # krwend1
    st.floats(min_value=0.1, max_value=5),  # now1
    st.floats(min_value=0.1, max_value=1),  # kroend1
    st.floats(min_value=0.1, max_value=5),  # nw2
    st.floats(min_value=0.1, max_value=1),  # krwend2
    st.floats(min_value=0.1, max_value=5),  # now2
    st.floats(min_value=0.1, max_value=1),  # kroend2
)
def test_normalize_nonlinpart_wo_hypo(
    swl,
    dswcr,
    dswlhigh,
    sorw,
    dsocrsorw,
    nw1,
    krwend1,
    now1,
    kroend1,
    nw2,
    krwend2,
    now2,
    kroend2,
):
    """Test the normalization code in utils.

    In particular the fill_value argument to scipy has been tuned to
    fulfill this code"""
    wo_low = WaterOil(swl=swl, swcr=swl + dswcr, sorw=sorw, socr=sorw + dsocrsorw)
    sorw_high = max(sorw - 0.01, 0)
    wo_high = WaterOil(
        swl=swl + dswlhigh,
        swcr=swl + dswlhigh + dswcr,
        sorw=sorw_high,
        socr=sorw_high + dsocrsorw,
    )
    wo_low.add_corey_water(nw=nw1, krwend=krwend1)
    wo_high.add_corey_water(nw=nw2, krwend=krwend2)
    wo_low.add_corey_oil(now=now1, kroend=kroend1)
    wo_high.add_corey_oil(now=now2, kroend=kroend2)

    krwn1, kron1 = normalize_nonlinpart_wo(wo_low)
    assert np.isclose(krwn1(0), 0)
    assert np.isclose(krwn1(1), krwend1)
    assert np.isclose(kron1(0), 0)
    assert np.isclose(kron1(1), kroend1)

    krwn2, kron2 = normalize_nonlinpart_wo(wo_high)
    assert np.isclose(krwn2(0), 0)
    assert np.isclose(krwn2(1), krwend2)
    assert np.isclose(kron2(0), 0)
    assert np.isclose(kron2(1), kroend2)


@given(
    st.floats(min_value=0, max_value=0.3),  # swirr
    st.floats(min_value=0.01, max_value=0.3),  # dswl
)
def test_normalize_pc(swirr, dswl):
    """Test that we can normalize a pc curve"""
    wateroil = WaterOil(swirr=swirr, swl=swirr + dswl)
    wateroil.add_simple_J()
    pc_max = wateroil.table["PC"].max()
    pc_min = wateroil.table["PC"].min()

    pc_fn = normalize_pc(wateroil)
    assert np.isclose(pc_fn(0), pc_max)
    assert np.isclose(pc_fn(1), pc_min)


def test_normalize_pc_wrongobject():
    """Test error message when wrong object is provided"""
    with pytest.raises(ValueError, match="Only WaterOil or GasOil"):
        normalize_pc({})


def test_normalize_emptypc():
    """Test that we can normalize both
    when pc is missing, and when it is all zero"""
    wateroil = WaterOil()
    pc_fn = normalize_pc(wateroil)
    assert np.isclose(pc_fn(0), 0)
    assert np.isclose(pc_fn(1), 0)

    wateroil = WaterOil(swl=0.01)
    wateroil.add_simple_J(g=0)
    pc_fn = normalize_pc(wateroil)
    assert np.isclose(pc_fn(0), 0)
    assert np.isclose(pc_fn(1), 0)


def test_normalize_nonlinpart_wo():
    """Manual tests for normalize_nonlinpart_wo"""
    wateroil = WaterOil(swl=0.1, swcr=0.12, sorw=0.05, h=0.05)
    wateroil.add_corey_water(nw=2.1, krwend=0.9)
    wateroil.add_corey_oil(now=3, kroend=0.8)
    krwn, kron = normalize_nonlinpart_wo(wateroil)

    assert np.isclose(krwn(0), 0)
    assert np.isclose(krwn(1), 0.9)

    # kron is normalized on son
    assert np.isclose(kron(0), 0)
    assert np.isclose(kron(1), 0.8)

    # Test with tricky endpoints
    h = 0.01
    wateroil = WaterOil(swl=h, swcr=h, sorw=h, h=h)
    wateroil.add_corey_water(nw=2.1, krwend=0.9)
    wateroil.add_corey_oil(now=3, kroend=0.8)
    krwn, kron = normalize_nonlinpart_wo(wateroil)
    assert np.isclose(krwn(0), 0.0)
    assert np.isclose(krwn(1), 0.9)
    assert np.isclose(kron(0), 0)
    assert np.isclose(kron(1), 0.8)

    # Test again with zero endpoints:
    wateroil = WaterOil(swl=0, swcr=0, sorw=0, h=0.01)
    wateroil.add_corey_water(nw=2.1, krwend=0.9)
    wateroil.add_corey_oil(now=3, kroend=0.8)
    krwn, kron = normalize_nonlinpart_wo(wateroil)
    assert np.isclose(krwn(0), 0.0)
    assert np.isclose(krwn(1), 0.9)
    assert np.isclose(kron(0), 0)
    assert np.isclose(kron(1), 0.8)

    # Test when endpoints are messed up:
    wateroil = WaterOil(swl=0.1, swcr=0.2, sorw=0.1, h=0.1)
    wateroil.add_corey_water(nw=2.1, krwend=0.6)
    wateroil.add_corey_oil(now=3, kroend=0.8)
    wateroil.swl = 0
    wateroil.swcr = 0
    wateroil.sorw = 0
    krwn, kron = normalize_nonlinpart_wo(wateroil)
    # These go well still, since we are at zero
    assert np.isclose(krwn(0), 0.0)
    assert np.isclose(kron(0), 0)
    # These do not match when endpoints are wrong
    assert not np.isclose(krwn(1), 0.6)
    # This is ok, as kro is anchored at swl
    assert np.isclose(kron(1), 0.8)

    # So fix endpoints!
    wateroil.swl = wateroil.table["SW"].min()
    wateroil.swcr = wateroil.estimate_swcr()
    wateroil.sorw = wateroil.estimate_sorw()
    # Try again
    krwn, kron = normalize_nonlinpart_wo(wateroil)
    assert np.isclose(krwn(0), 0.0)
    assert np.isclose(kron(0), 0)
    assert np.isclose(krwn(1), 0.6)
    assert np.isclose(kron(1), 0.8)


def test_tag_preservation():
    """Test that we can preserve tags/comments through interpolation"""
    wo_low = WaterOil(swl=0.1)
    wo_high = WaterOil(swl=0.2)
    wo_low.add_corey_water(nw=2)
    wo_high.add_corey_water(nw=3)
    wo_low.add_corey_oil(now=2)
    wo_high.add_corey_oil(now=3)
    interpolant1 = interpolate_wo(wo_low, wo_high, parameter=0.1, h=0.2)
    assert "Interpolated to 0.1" in interpolant1.tag
    sat_table_str_ok(interpolant1.SWOF())

    wo_high.tag = "FOOBAR"
    interpolant2 = interpolate_wo(wo_low, wo_high, parameter=0.1, h=0.2)
    assert "Interpolated to 0.1" in interpolant2.tag
    assert "between" in interpolant2.tag
    assert wo_high.tag in interpolant2.tag
    sat_table_str_ok(interpolant2.SWOF())
    # wo_low.tag was empty deliberately here.

    # When wo_log and wo_high has the same tag:
    wo_low.tag = "FOOBAR"
    interpolant3 = interpolate_wo(wo_low, wo_high, parameter=0.1, h=0.2)
    assert "Interpolated to 0.1" in interpolant3.tag
    assert "between" not in interpolant3.tag
    assert wo_high.tag in interpolant3.tag
    sat_table_str_ok(interpolant3.SWOF())

    # Explicit tag:
    interpolant4 = interpolate_wo(
        wo_low, wo_high, parameter=0.1, h=0.2, tag="Explicit tag"
    )
    assert interpolant4.tag == "Explicit tag"

    # Tag with newline
    interpolant6 = interpolate_wo(
        wo_low, wo_high, parameter=0.1, h=0.2, tag="Explicit tag\non two lines"
    )
    assert "Explicit tag" in interpolant6.tag
    print(interpolant6.SWOF())
    sat_table_str_ok(interpolant6.SWOF())

    # Empty tag:
    interpolant5 = interpolate_wo(wo_low, wo_high, parameter=0.1, h=0.2, tag="")
    assert interpolant5.tag == ""

    # Also sample check for GasOil (calls the same code)
    go_low = GasOil()
    go_high = GasOil()
    go_low.add_corey_gas(ng=2)
    go_high.add_corey_gas(ng=3)
    go_low.add_corey_oil(nog=2)
    go_high.add_corey_oil(nog=3)
    interpolant1 = interpolate_go(go_low, go_high, parameter=0.1, h=0.2)
    assert "Interpolated to 0.1" in interpolant1.tag
    sat_table_str_ok(interpolant1.SGOF())


@slow_hypothesis
@given(
    st.floats(min_value=0.01, max_value=0.1),  # swl
    st.floats(min_value=0, max_value=0.0),  # dswcr
    st.floats(min_value=0, max_value=0.1),  # dswlhigh
    st.floats(min_value=0, max_value=0.3),  # sorw
    st.floats(min_value=0, max_value=0.3),  # dsocrsorw
    st.floats(min_value=1, max_value=5),  # nw_l
    st.floats(min_value=1, max_value=5),  # nw_h
    st.floats(min_value=1, max_value=5),  # now_l
    st.floats(min_value=1, max_value=5),  # now_h
    st.floats(min_value=0.1, max_value=1),  # krwend_l
    st.floats(min_value=0.1, max_value=1),  # krwend_h
    st.floats(min_value=0.1, max_value=1),  # kroend_l
    st.floats(min_value=0.1, max_value=1),  # kroend_h
)
def test_interpolate_wo(
    swl,
    dswcr,
    dswlhigh,
    sorw,
    dsocrsorw,
    nw_l,
    nw_h,
    now_l,
    now_h,
    krwend_l,
    krwend_h,
    kroend_l,
    kroend_h,
):
    """
    Generate two random WaterOil curves, interpolate between them
    and check that the difference between each interpolant is small,
    this essentially checks that we can go continously between the
    two functions.
    """
    wo_low = WaterOil(swl=swl, swcr=swl + dswcr, sorw=sorw, socr=sorw + dsocrsorw)
    sorw_high = max(sorw - 0.01, 0)
    wo_high = WaterOil(
        swl=swl + dswlhigh,
        swcr=swl + dswlhigh + dswcr,
        sorw=max(sorw - 0.01, 0),
        socr=sorw_high + dsocrsorw,
    )
    wo_low.add_corey_water(nw=nw_l, krwend=krwend_l)
    wo_high.add_corey_water(nw=nw_h, krwend=krwend_h)
    wo_low.add_corey_oil(now=now_l, kroend=kroend_l)
    wo_high.add_corey_oil(now=now_h, kroend=kroend_h)
    ips = []
    ip_dist = 0.05
    for tparam in np.arange(0, 1 + ip_dist, ip_dist):
        wo_ip = interpolate_wo(wo_low, wo_high, tparam)
        check_table(wo_ip.table)
        assert wo_ip.tag
        ips.append(wo_ip)
        assert 0 < wo_ip.crosspoint() < 1

    # Distances between low and interpolants:
    dists = [
        (wo_low.table - interp.table)[["KRW", "KROW"]].sum().sum() for interp in ips
    ]

    # wo_low should be reproduced exactly:
    assert np.isclose(dists[0], 0)

    # Distance between high and the last interpolant
    assert (wo_high.table - ips[-1].table)[["KRW", "KROW"]].sum().sum() < 0.01

    # Distances between low and interpolants:
    dists = [
        (wo_low.table - interp.table)[["KRW", "KROW"]].sum().sum() for interp in ips
    ]
    print(
        f"Interpolation, mean: {np.mean(dists)}, min: {min(dists)}, "
        f"max: {max(dists)}, std: {np.std(np.diff(dists[1:]))} ip-par-dist: {ip_dist}"
    )
    # All curves that are close in parameter t, should be close in sum().sum().
    # That means that diff of the distances should be similar,
    # that is the std.dev of the distances is low:
    ip_dist_std = np.std(np.diff(dists[1:]))  # This number depends on 'h' and 't' range
    # (avoiding the first which reproduces wo_low
    if ip_dist_std > 1.2:  # Found by trial and error
        print(f"ip_dist_std: {ip_dist_std}")
        print(dists)

        _, mpl_ax = plt.subplots()
        wo_low.plotkrwkrow(mpl_ax=mpl_ax, color="red")
        wo_high.plotkrwkrow(mpl_ax=mpl_ax, color="blue")
        for interp in ips:
            interp.plotkrwkrow(mpl_ax=mpl_ax, color="green")
        plt.show()
        assert False

    # If this fails, there is a combination of parameters that might
    # trigger a discontinuity in the interpolants, which we don't want.


@slow_hypothesis
@given(
    st.floats(min_value=0.01, max_value=0.1),  # swl
    st.floats(min_value=0, max_value=0.0),  # dswcr
    st.floats(min_value=0, max_value=0.1),  # dswlhigh
    st.floats(min_value=0, max_value=0.3),  # sorw
    st.floats(min_value=0.1, max_value=2),  # a_l
    st.floats(min_value=0.1, max_value=2),  # a_h
    st.floats(min_value=-2, max_value=-0.1),  # b_l
    st.floats(min_value=-2, max_value=-0.5),  # b_l
)
def test_interpolate_wo_pc(swl, dswcr, dswlhigh, sorw, a_l, a_h, b_l, b_h):
    """
    Generate two random WaterOil curves, interpolate pc between them
    and check that the difference between each interpolant is small,
    this essentially checks that we can go continously between the
    two functions.
    """
    wo_low = WaterOil(swl=swl, swcr=swl + dswcr, sorw=sorw)
    wo_high = WaterOil(
        swl=swl + dswlhigh, swcr=swl + dswlhigh + dswcr, sorw=max(sorw - 0.01, 0)
    )
    wo_low.add_corey_water()
    wo_high.add_corey_water()
    wo_low.add_corey_oil()
    wo_high.add_corey_oil()
    wo_low.add_simple_J(a=a_l, b=b_l)
    wo_high.add_simple_J(a=a_h, b=b_h)
    ips = []
    ip_dist = 0.05
    for t in np.arange(0, 1 + ip_dist, ip_dist):
        wo_ip = interpolate_wo(wo_low, wo_high, t)
        check_table(wo_ip.table)
        ips.append(wo_ip)
        assert 0 < wo_ip.crosspoint() < 1

    # Distances between low and interpolants:
    dists = [(wo_low.table - interp.table)[["PC"]].sum().sum() for interp in ips]
    assert np.isclose(dists[0], 0)

    # Distance between high and the last interpolant
    assert (wo_high.table - ips[-1].table)[["PC"]].sum().sum() < 0.01

    # Distances between low and interpolants:
    dists = [(wo_low.table - interp.table)[["PC"]].sum().sum() for interp in ips]
    print(
        f"Interpolation, mean: {np.mean(dists)}, min: {min(dists)}, "
        f"max: {max(dists)}, std: {np.std(np.diff(dists[1:]))} ip-par-dist: {ip_dist}"
    )
    assert np.isclose(dists[0], 0)  # Reproducing wo_low
    # All curves that are close in parameter t, should be close in sum().sum().
    # That means that diff of the distances should be similar,
    # that is the std.dev of the distances is low:
    ip_dist_std = np.std(np.diff(dists[1:]))  # This number depends on 'h' and 't' range
    # (avoiding the first which reproduces go_low
    if ip_dist_std > 1.0:  # Found by trial and error
        print(f"ip_dist_std: {ip_dist_std}")
        print(dists)
        _, mpl_ax = plt.subplots()
        wo_low.plotpc(mpl_ax=mpl_ax, color="red", logyscale=True)
        wo_high.plotpc(mpl_ax=mpl_ax, color="blue", logyscale=True)
        for interp in ips:
            interp.plotpc(mpl_ax=mpl_ax, color="green", logyscale=True)
        plt.show()
        assert False

    # If this fails, there is a combination of parameters that might
    # trigger a discontinuity in the interpolants, which we don't want.


@slow_hypothesis
@given(
    st.floats(min_value=0, max_value=0.1),  # swl
    st.floats(min_value=0, max_value=0.1),  # sgcr
    st.floats(min_value=0, max_value=0.1),  # dswlhigh
    st.floats(min_value=0, max_value=0.1),  # dsgcr
    st.floats(min_value=0, max_value=0.3),  # sorg
    st.floats(min_value=0, max_value=0.1),  # dsorg
    st.booleans(),  # sgrononzero
    st.floats(min_value=0.1, max_value=5),  # ng1
    st.floats(min_value=0.01, max_value=1),  # krgend1
    st.floats(min_value=0.1, max_value=5),  # nog1
    st.floats(min_value=0.01, max_value=1),  # kroend1
    st.floats(min_value=0.01, max_value=1),  # kromax1
    st.floats(min_value=0.1, max_value=5),  # ng2
    st.floats(min_value=0.01, max_value=1),  # krgend2
    st.floats(min_value=0.1, max_value=5),  # nog2
    st.floats(min_value=0.01, max_value=1),  # kroend2
    st.floats(min_value=0.01, max_value=1),  # kromax2
)
def test_normalize_nonlinpart_go_hypo(
    swl,
    sgcr,
    dswlhigh,
    dsgcr,
    sorg,
    dsorg,
    sgrononzero,
    ng1,
    krgend1,
    nog1,
    kroend1,
    kromax1,
    ng2,
    krgend2,
    nog2,
    kroend2,
    kromax2,
):
    """Test the normalization code in utils.

    In particular the fill_value argument to scipy has been tuned to
    fulfill this code"""
    kroend1 = min(kroend1, kromax1)
    kroend2 = min(kroend2, kromax2)

    if sgrononzero:
        sgro_low = sgcr
        sgro_high = sgcr + dsgcr
    else:
        sgro_low = 0.0
        sgro_high = 0.0

    go_low = GasOil(swl=swl, sgcr=sgcr, sorg=sorg, sgro=sgro_low)
    go_high = GasOil(
        swl=swl + dswlhigh,
        sgcr=sgcr + dsgcr,
        sgro=sgro_high,
        sorg=max(sorg - dsorg, 0),
    )
    go_low.add_corey_gas(ng=ng1, krgend=krgend1)
    go_high.add_corey_gas(ng=ng2, krgend=krgend2)
    go_low.add_corey_oil(nog=nog1, kroend=kroend1, kromax=kromax1)
    go_high.add_corey_oil(nog=nog2, kroend=kroend2, kromax=kromax2)

    krgn1, kron1 = normalize_nonlinpart_go(go_low)
    assert np.isclose(krgn1(0), 0)
    assert np.isclose(krgn1(1), krgend1)
    assert np.isclose(kron1(0), 0)
    assert np.isclose(kron1(1), kroend1)

    krgn2, kron2 = normalize_nonlinpart_go(go_high)
    assert np.isclose(krgn2(0), 0)
    assert np.isclose(krgn2(1), krgend2)
    assert np.isclose(kron2(0), 0)
    assert np.isclose(kron2(1), kroend2)


def test_normalize_nonlinpart_go():
    """Manual tests for normalize_nonlinpart_go"""
    gasoil = GasOil(swl=0.1, sgcr=0.12, sgro=0.12, sorg=0.05, h=0.05)
    gasoil.add_corey_gas(ng=2.1, krgend=0.9)
    gasoil.add_corey_oil(nog=3, kroend=0.75, kromax=0.8)
    krgn, kron = normalize_nonlinpart_go(gasoil)

    assert np.isclose(krgn(0), 0)
    assert np.isclose(krgn(1), 0.9)

    # kron is normalized on son
    assert np.isclose(kron(0), 0)
    assert np.isclose(kron(1), 0.75)

    # Test with tricky endpoints
    h = 0.01
    gasoil = GasOil(swl=h, sgcr=h, sorg=h, h=h)
    gasoil.add_corey_gas(ng=2.1, krgend=0.9)
    gasoil.add_corey_oil(nog=3, kroend=0.8)
    krgn, kron = normalize_nonlinpart_go(gasoil)
    assert np.isclose(krgn(0), 0.0)
    assert np.isclose(krgn(1), 0.9)
    assert np.isclose(kron(0), 0)
    assert np.isclose(kron(1), 0.8)

    # Test again with zero endpoints:
    gasoil = GasOil(swl=0, sgcr=0, sorg=0, h=0.01)
    gasoil.add_corey_gas(ng=2.1, krgend=0.9)
    gasoil.add_corey_oil(nog=3, kroend=0.8)
    krgn, kron = normalize_nonlinpart_go(gasoil)
    assert np.isclose(krgn(0), 0.0)
    assert np.isclose(krgn(1), 0.9)
    assert np.isclose(kron(0), 0)
    assert np.isclose(kron(1), 0.8)

    # Test when endpoints are messed up (cleared)
    gasoil = GasOil(swl=0.1, sgcr=0.2, sorg=0.1, h=0.1)
    gasoil.add_corey_gas(ng=2.1, krgend=0.6)
    gasoil.add_corey_oil(nog=3, kroend=0.8)
    gasoil.swl = 0
    gasoil.sgcr = 0
    gasoil.sorg = 0
    krgn, kron = normalize_nonlinpart_go(gasoil)
    # These go well still, since we are at zero
    assert np.isclose(krgn(0), 0.0)
    assert np.isclose(kron(0), 0)
    assert np.isclose(kron(1), 0.8)
    # These do not match when endpoints are wrong
    assert not np.isclose(krgn(1), 0.6)

    # So fix endpoints!
    gasoil.swl = 1 - gasoil.table["SG"].max()
    gasoil.sgcr = gasoil.estimate_sgcr()
    gasoil.sorg = gasoil.estimate_sorg()
    # Try again
    krgn, kron = normalize_nonlinpart_go(gasoil)
    assert np.isclose(krgn(0), 0.0)
    assert np.isclose(kron(0), 0)
    assert np.isclose(krgn(1), 0.6)
    assert np.isclose(kron(1), 0.8)


def test_ip_wo_kroend():
    """Test behaviour of kroend under interpolation"""
    wo_low = WaterOil(swl=0, swcr=0.1, sorw=0.2)
    wo_low.add_corey_water(nw=2, krwend=0.5, krwmax=0.7)
    wo_low.add_corey_oil(now=2, kroend=0.6)

    wo_high = WaterOil(swl=0.02, swcr=0.05, sorw=0.1)
    wo_high.add_corey_water(nw=2, krwend=0.5, krwmax=0.72)
    wo_high.add_corey_oil(now=2, kroend=0.7)

    # Interpolate to midpoint between the curves above
    wo_ip = interpolate_wo(wo_low, wo_high, 0.5)

    # kroend at mean swl:
    assert float_df_checker(wo_ip.table, "SW", 0.01, "KROW", (0.6 + 0.7) / 2.0)

    assert float_df_checker(wo_ip.table, "SW", 1, "KRW", 0.71)
    assert float_df_checker(wo_ip.table, "SW", 1 - 0.15, "KRW", 0.5)


def test_ip_go_kroendmax():
    """Test behaviour of kroend/kromax under interpolation, gas condensate modelling"""
    go_low = GasOil(swl=0, sgro=0.1, sgcr=0.1)
    go_high = GasOil(swl=0, sgro=0)
    go_low.add_corey_gas()
    go_low.add_corey_oil(nog=2, kroend=0.5, kromax=1)
    go_high.add_corey_gas()
    go_high.add_corey_oil(nog=2, kroend=1)

    # Interpolate to midpoint between the curves above
    go_ip = interpolate_go(go_low, go_high, 0.5)

    # kro(sg=0) is 1 for all interpolants:
    assert float_df_checker(go_ip.table, "SG", 0.0, "KROG", 1.0)

    # kro(sg=mean(sgro)) = mean kroeend
    assert float_df_checker(go_ip.table, "SG", (0 + 0.1) / 2.0, "KROG", 0.75)

    assert np.isclose(go_ip.estimate_sgro(), (0 + 0.1) / 2.0)


def test_ip_go_kroend():
    """Test behaviour of kroend under interpolation"""
    go_low = GasOil(swl=0, sgcr=0.1, sorg=0.2)
    go_low.add_corey_gas(ng=2, krgend=0.5, krgmax=0.7)
    go_low.add_corey_oil(nog=2, kroend=0.6)

    go_high = GasOil(swl=0.02, sgcr=0.05, sorg=0.1)
    go_high.add_corey_gas(ng=2, krgend=0.5, krgmax=0.72)
    go_high.add_corey_oil(nog=2, kroend=0.7)

    # Interpolate to midpoint between the curves above
    go_ip = interpolate_go(go_low, go_high, 0.5)

    # kroend at sg zero:
    assert float_df_checker(go_ip.table, "SG", 0.0, "KROG", (0.6 + 0.7) / 2.0)

    assert np.isclose(go_ip.swl, 0.01)
    assert np.isclose(go_ip.sorg, 0.15)

    # krgmax at 1 - swl:
    assert float_df_checker(go_ip.table, "SG", 1 - go_ip.swl, "KRG", 0.71)
    # krgend at 1 - swl - sorg
    assert float_df_checker(go_ip.table, "SG", 1 - go_ip.swl - go_ip.sorg, "KRG", 0.5)

    # If krgendanchor is None, then krgmax should be irrelevant
    go_low = GasOil(swl=0, sgcr=0.1, sorg=0.2, krgendanchor="")
    go_low.add_corey_gas(ng=5, krgend=0.5, krgmax=0.7)
    # krgmax will trigger warning message, intended, as the 0.7
    # value here will mean nothing.
    go_low.add_corey_oil(nog=2, kroend=0.6)

    go_high = GasOil(swl=0.02, sgcr=0.05, sorg=0.1, krgendanchor="")
    go_high.add_corey_gas(ng=4, krgend=0.5, krgmax=0.72)
    # krgmax will trigger warning message, intended.
    go_high.add_corey_oil(nog=2, kroend=0.7)

    # Interpolate to midpoint between the curves above
    go_ip = interpolate_go(go_low, go_high, 0.5, h=0.1)

    assert float_df_checker(go_ip.table, "SG", 0.0, "KROG", (0.6 + 0.7) / 2.0)

    # Activate these line to see a bug, interpolation_go
    # does not honor krgendanchorA:
    # _, mpl_ax = plt.subplots()
    # go_low.plotkrgkrog(mpl_ax=mpl_ax, color="red")
    # go_high.plotkrgkrog(mpl_ax=mpl_ax, color="blue")
    # go_ip.plotkrgkrog(mpl_ax=mpl_ax, color="green")
    # plt.show()

    # krgmax is irrelevant, krgend is used here:
    assert float_df_checker(go_ip.table, "SG", 1 - 0.01, "KRG", 0.5)

    # Also check that estimated new sgcr is between the inputs:
    assert 0.05 <= go_ip.estimate_sgcr() <= 0.1

    # Do we get into trouble if krgendanchor is different in low and high?
    go_low = GasOil(swl=0, sgcr=0.1, sorg=0.2, krgendanchor="sorg")
    go_low.add_corey_gas(ng=2, krgend=0.5, krgmax=0.7)
    go_low.add_corey_oil(nog=2, kroend=0.6)

    go_high = GasOil(swl=0.02, sgcr=0.05, sorg=0.1, krgendanchor="")
    go_high.add_corey_gas(ng=2, krgend=0.5)
    go_high.add_corey_oil(nog=2, kroend=0.7)

    # Interpolate to midpoint between the curves above
    go_ip = interpolate_go(go_low, go_high, 0.5)

    assert float_df_checker(go_ip.table, "SG", 0.0, "KROG", (0.6 + 0.7) / 2.0)

    # max(krg) is here avg of krgmax and krgend from the differnt tables:
    assert float_df_checker(go_ip.table, "SG", 1 - 0.01, "KRG", 0.6)

    # krgend at 1 - swl - sorg, non-trivial expression, so a numerical
    # value is used here in the test:
    assert float_df_checker(go_ip.table, "SG", 1 - 0.01 - 0.15, "KRG", 0.4491271)

    # krog-zero at 1 - swl - sorg:
    assert float_df_checker(go_ip.table, "SG", 1 - 0.01 - 0.15, "KROG", 0)


@slow_hypothesis
@given(
    st.floats(min_value=0, max_value=0.1),  # swl
    st.floats(min_value=0, max_value=0.1),  # sgcr
    st.floats(min_value=0, max_value=0.1),  # dsgcr
    st.floats(min_value=0, max_value=0.1),  # dswlhigh
    st.floats(min_value=0, max_value=0.3),  # sorg
    st.floats(min_value=0, max_value=0.1),  # dsorg
    st.floats(min_value=1, max_value=5),  # ng_l
    st.floats(min_value=1, max_value=5),  # ng_h
    st.floats(min_value=1, max_value=5),  # nog_l
    st.floats(min_value=1, max_value=5),  # nog_h
    st.floats(min_value=0.1, max_value=1),  # krgend_l
    st.floats(min_value=0.1, max_value=1),  # krgend_h
    st.floats(min_value=0.1, max_value=1),  # kroend_l
    st.floats(min_value=0.1, max_value=1),  # kroend_h
)
def test_interpolate_go(
    swl,
    sgcr,
    dsgcr,
    dswlhigh,
    sorg,
    dsorg,
    ng_l,
    ng_h,
    nog_l,
    nog_h,
    krgend_l,
    krgend_h,
    kroend_l,
    kroend_h,
):
    """Test many possible combinations of interpolation between two
    Corey gasoil curves, looking for numerical corner cases"""
    h = 0.01
    go_low = GasOil(swl=swl, sgcr=sgcr, sorg=sorg, h=h)
    go_high = GasOil(
        swl=swl + dswlhigh, sgcr=sgcr + dsgcr, sorg=max(sorg - dsorg, 0), h=h
    )
    go_low.add_corey_gas(ng=ng_l, krgend=krgend_l)
    go_high.add_corey_gas(ng=ng_h, krgend=krgend_h)
    go_low.add_corey_oil(nog=nog_l, kroend=kroend_l)
    go_high.add_corey_oil(nog=nog_h, kroend=kroend_h)
    ips = []
    ip_dist = 0.05
    for t in np.arange(0, 1 + ip_dist, ip_dist):
        go_ip = interpolate_go(go_low, go_high, t)
        check_table(go_ip.table)
        ips.append(go_ip)
        assert 0 < go_ip.crosspoint() < 1

        # sgcr is non-trivial, if exponents are high, an effective sgcr might
        # be larger than the value used to initialize the curve. Try to be
        # permissive enough here. This can even cause the interpolant to be
        # outside the low-high envelope, but it is the way it is supposed to
        # be when sgcr is interpolated separately.
        sgcr_low = min(
            go_low.sgcr, go_low.estimate_sgcr(), go_high.sgcr, go_high.estimate_sgcr()
        )
        sgcr_high = max(
            go_low.sgcr, go_low.estimate_sgcr(), go_high.sgcr, go_high.estimate_sgcr()
        )

        sgcr_ip = go_ip.estimate_sgcr()

        sgcr_lower_bound_ok = sgcr_low - h - epsilon < sgcr_ip
        sgcr_upper_bound_ok = sgcr_ip < sgcr_high + h + epsilon

        assert sgcr_lower_bound_ok
        assert sgcr_upper_bound_ok

    # Distances between low and interpolants:
    dists = [
        (go_low.table - interp.table)[["KRG", "KROG"]].sum().sum() for interp in ips
    ]
    print(
        f"Interpolation, mean: {np.mean(dists)}, min: {min(dists)}, "
        f"max: {max(dists)}, std: {np.std(np.diff(dists[1:]))} ip-par-dist: {ip_dist}"
    )
    assert np.isclose(dists[0], 0)  # Reproducing go_low
    # All curves that are close in parameter t, should be close in sum().sum().
    # That means that diff of the distances should be similar,
    # that is the std.dev of the distances is low:
    ip_dist_std = np.std(
        np.diff(dists[1:])
    )  # This number depends on 'h' and 't' range, and
    # by how different the low and high is.
    # (avoiding the first which reproduces go_low
    if ip_dist_std > 1.0:  # number found from trial and error.
        print(f"ip_dist_std: {ip_dist_std}")
        print(dists)
        _, mpl_ax = plt.subplots()
        go_low.plotkrgkrog(mpl_ax=mpl_ax, color="red")
        go_high.plotkrgkrog(mpl_ax=mpl_ax, color="blue")
        for interp in ips:
            interp.plotkrgkrog(mpl_ax=mpl_ax, color="green")
        plt.show()
        assert False

    # If this fails, there is a combination of parameters that might
    # trigger a discontinuity in the interpolants, which we don't want.


def test_interpolations_go_fromtable():
    """Test based on bug exposed in pyscal 0.6.1, where sgcr
    was underestimated in interpolations following add_fromtable().
    """
    base = pd.DataFrame(
        columns=["SG", "KRG", "KROG"],
        data=[
            [0.0, 0.0, 1.0],
            [0.1, 0.0, 1.0],
            [0.2, 0.0, 1.0],  # sgcr
            [0.3, 0.1, 0.9],
            [0.8, 0.8, 0.0],  # sorg
            [0.9, 0.9, 0.0],
            [1.0, 1.0, 0.0],
        ],
    )
    opt = pd.DataFrame(
        columns=["SG", "KRG", "KROG"],
        data=[
            [0.0, 0.0, 1.0],
            [0.1, 0.0, 1.0],
            [0.3, 0.0, 1.0],
            [0.4, 0.1, 0.2],  # sgcr
            [0.9, 0.9, 0.0],  # sorg
            [0.95, 0.95, 0.0],
            [1.0, 1.0, 0.0],
        ],
    )
    go_base = GasOil(h=0.01)
    go_base.add_fromtable(base)
    assert np.isclose(go_base.estimate_sgcr(), 0.2)
    assert np.isclose(go_base.estimate_sorg(), 0.2)
    go_opt = GasOil(h=0.01)
    go_opt.add_fromtable(opt)
    assert np.isclose(go_opt.estimate_sgcr(), 0.3)
    assert np.isclose(go_opt.estimate_sorg(), 0.1)

    go_ip = interpolate_go(go_base, go_opt, 0.5, h=0.01)
    assert np.isclose(go_ip.estimate_sgcr(), 0.25)
    assert np.isclose(go_ip.estimate_sorg(), 0.15)


def test_interpolations_wo_fromtable():
    """Analog test as test_interpolations_go_fromtable().

    Pyscal 0.6.1 and earlier fails this test sorw.
    """
    base = pd.DataFrame(
        columns=["SW", "KRW", "KROW"],
        data=[
            [0.0, 0.0, 1.0],
            [0.1, 0.0, 1.0],
            [0.2, 0.0, 1.0],  # swcr
            [0.3, 0.1, 0.9],
            [0.8, 0.8, 0.0],  # sorw
            [0.9, 0.9, 0.0],
            [1.0, 1.0, 0.0],
        ],
    )
    opt = pd.DataFrame(
        columns=["SW", "KRW", "KROW"],
        data=[
            [0.0, 0.0, 1.0],
            [0.1, 0.0, 1.0],
            [0.3, 0.0, 1.0],
            [0.4, 0.1, 0.2],  # swcr
            [0.9, 0.9, 0.0],  # sorw
            [0.95, 0.95, 0.0],
            [1.0, 1.0, 0.0],
        ],
    )
    wo_base = WaterOil(h=0.01)
    wo_base.add_fromtable(base)
    assert np.isclose(wo_base.estimate_swcr(), 0.2)
    assert np.isclose(wo_base.estimate_sorw(), 0.2)
    wo_opt = WaterOil(h=0.01)
    wo_opt.add_fromtable(opt)
    assert np.isclose(wo_opt.estimate_swcr(), 0.3)
    assert np.isclose(wo_opt.estimate_sorw(), 0.1)

    wo_ip = interpolate_wo(wo_base, wo_opt, 0.5, h=0.01)
    assert np.isclose(wo_ip.estimate_swcr(), 0.25)
    assert np.isclose(wo_ip.estimate_sorw(), 0.15)


def test_interpolations_wo_fromtable_socr():
    """Test that socr is correctly handled through
    both add_fromtable() and interpolation."""
    base = pd.DataFrame(
        columns=["SW", "KRW", "KROW"],
        data=[
            [0.0, 0.0, 1.0],
            [0.1, 0.0, 1.0],
            [0.2, 0.0, 1.0],  # swcr
            [0.3, 0.1, 0.9],
            [0.7, 0.3, 0.0],  # socr
            [0.8, 0.8, 0.0],  # sorw
            [0.9, 0.9, 0.0],
            [1.0, 1.0, 0.0],
        ],
    )
    opt = pd.DataFrame(
        columns=["SW", "KRW", "KROW"],
        data=[
            [0.0, 0.0, 1.0],
            [0.1, 0.0, 1.0],
            [0.3, 0.0, 1.0],
            [0.4, 0.1, 0.2],  # swcr
            [0.9, 0.9, 0.0],  # sorw == socr
            [0.95, 0.95, 0.0],
            [1.0, 1.0, 0.0],
        ],
    )
    wo_base = WaterOil(h=0.01)
    wo_base.add_fromtable(base)
    assert np.isclose(wo_base.estimate_swcr(), 0.2)
    assert np.isclose(wo_base.estimate_sorw(), 0.2)
    assert np.isclose(wo_base.estimate_socr(), 0.3)
    wo_opt = WaterOil(h=0.01)
    wo_opt.add_fromtable(opt)
    assert np.isclose(wo_opt.estimate_swcr(), 0.3)
    assert np.isclose(wo_opt.estimate_sorw(), 0.1)
    assert np.isclose(wo_opt.estimate_socr(), 0.1)

    wo_ip = interpolate_wo(wo_base, wo_opt, 0.5, h=0.01)
    assert np.isclose(wo_ip.estimate_swcr(), 0.25)
    assert np.isclose(wo_ip.estimate_sorw(), 0.15)
    assert np.isclose(wo_ip.estimate_socr(), 0.2)


@given(
    st.floats(min_value=0, max_value=0.1),  # sgro_low
    st.floats(min_value=0, max_value=0.1),  # sgro_high
)
def test_gascond_interpolation(sgro_low, sgro_high):
    """sgro is required to be either 0 or sgcr, and interpolations
    will crash when this is not the case. This test validates that
    we can let sgro and sgcr go to zero, and that we always are able
    to interpolate without crashes."""
    go_low = GasOil(sgro=sgro_low, sgcr=sgro_low)
    go_high = GasOil(sgro=sgro_high, sgcr=sgro_high)
    go_low.add_corey_gas()
    go_low.add_corey_oil()
    go_high.add_corey_gas()
    go_high.add_corey_oil()

    go_ip = interpolate_go(go_low, go_high, parameter=0.5)
    check_table(go_ip.table)


def test_gasoil_gascond_fails():
    """Interpolation between an object for gas condendensate (sgro > 0)
    and an object with sgro = 0 (regular gas-oil) should fail"""
    gascond = GasOil(sgro=0.1, sgcr=0.1)
    gasoil = GasOil(sgro=0.0, sgcr=0.1)

    gasoil.add_corey_gas()
    gasoil.add_corey_oil()
    gascond.add_corey_gas()
    gascond.add_corey_oil()

    # interpolation parameter at zero is no interpolation, just lookup,
    # so it works fine:
    check_table(interpolate_go(gasoil, gascond, parameter=0.0).table)
    check_table(interpolate_go(gasoil, gascond, parameter=epsilon).table)
    check_table(interpolate_go(gasoil, gascond, parameter=10 * epsilon).table)

    with pytest.raises(ValueError, match="Interpolated sgro"):
        interpolate_go(gasoil, gascond, parameter=100 * epsilon)

    with pytest.raises(ValueError, match="Interpolated sgro"):
        interpolate_go(gasoil, gascond, parameter=0.5)

    check_table(interpolate_go(gasoil, gascond, parameter=1.0 - epsilon).table)
    check_table(interpolate_go(gasoil, gascond, parameter=1.0).table)


def test_endpointpitfalls_in_interpolation_socr():
    """Demonstration of behaviour and pitfalls around endpoints (socr in particular)
    when initializing objects through discretization (add_fromtable()) and how it will
    affect interpolation.
    """
    wo_low = WaterOil(swl=0.1, sorw=0.12)
    wo_low.add_LET_water(l=2, e=2, t=1)
    wo_low.add_LET_oil(l=5, e=5, t=1)

    assert wo_low.socr == wo_low.sorw  # By initialization

    # The KROW values approach zero faster than KRW approaches 1,
    # which makes the tabulated values indicate socr > sorw.
    # This is correct behaviour given finite floating point accuracy.
    assert wo_low.estimate_socr() > wo_low.estimate_sorw()  # difference is around 0.01
    assert np.isclose(wo_low.estimate_sorw(), 0.12)  # correct estimate

    # A not-so-extreme curve-set
    wo_high = WaterOil(swl=0.1, sorw=0.12)
    wo_high.add_LET_water(l=1, e=2, t=1)
    wo_high.add_LET_oil(l=1, e=2, t=1)
    assert wo_low.socr == wo_low.sorw
    # No surprises:
    assert np.isclose(wo_high.estimate_socr(), wo_high.estimate_sorw())

    # If the original wo_low object is then reinitialized through its own
    # discretization (e.g. via a SWOF file on disk and then back again, or via
    # its table property and then add_fromtable()), the original knowledge that
    # socr was defaulted and thus equal to sorw is lost and the best pyscal can
    # do is to use the estimators
    wo_low_dataframe = wo_low.table.copy()
    wo_low_fromtable = WaterOil(wo_low_dataframe["SW"].min())
    wo_low_fromtable.add_fromtable(wo_low_dataframe)
    # The properties of the WaterOil object are now the estimated values.
    assert wo_low_fromtable.socr > wo_low_fromtable.sorw

    # Interpolation from the original low object or the discretized low object
    # will be slightly different:
    wo_ip_original = interpolate_wo(wo_low, wo_high, 0.5)
    wo_ip_throughdiscretization = interpolate_wo(wo_low_fromtable, wo_high, 0.5)
    # The right end of the curves will have slightly lower values due to
    # the interpreteded socr for wo_ip_throughdiscretization:
    assert (
        wo_ip_original.table["KROW"].tail(20).sum()
        > wo_ip_throughdiscretization.table["KROW"].tail(20).sum()
    )
    # But it can be circumvented by explicitly stating the endpoints when
    # using add_fromtable()
    wo_low_fromtable_fixed = WaterOil(wo_low_dataframe["SW"].min())
    wo_low_fromtable_fixed.add_fromtable(wo_low_dataframe, sorw=0.12, socr=0.12)
    wo_ip_throughdiscretization_fixed = interpolate_wo(
        wo_low_fromtable_fixed, wo_high, 0.5
    )
    assert np.isclose(
        wo_ip_original.table["KROW"].tail(20).sum(),
        wo_ip_throughdiscretization_fixed.table["KROW"].tail(20).sum(),
    )

    # Plotting the minor discrepancy (zoom in to see the differences)
    # _, mpl_ax = pyplot.subplots()
    # wo_ip_original.plotkrwkrow(mpl_ax, color="green", alpha=0.6)
    # wo_ip_throughdiscretization.plotkrwkrow(mpl_ax, color="red", alpha=0.6)
    # pyplot.show()



"""Test module for monotonicity support functions in pyscal"""

import os

import numpy as np
import pandas as pd
import pytest

from pyscal.utils.monotonicity import (
    check_almost_monotone,
    check_limits,
    clip_accumulate,
    modify_dframe_monotonicity,
    rows_to_be_fixed,
    validate_monotonicity_arg,
)
from pyscal.utils.string import df2str


def test_df2str_monotone():
    """Test the monotonicity enforcement in df2str()

    This test function essentially tests the function
    utils/monotonicity.py::modify_dframe_monotonicity
    """

    # A constant nonzero column, makes no sense as capillary pressure
    # but still we ensure it runs in eclipse:
    assert (
        df2str(pd.DataFrame(data=[1, 1, 1]), digits=2, monotonicity={0: {"sign": -1}})
        == os.linesep.join(["1.00", "0.99", "0.98"]) + os.linesep
    )
    assert (
        df2str(pd.DataFrame(data=[1, 1, 1]), digits=2, monotonicity={0: {"sign": -1}})
        == os.linesep.join(["1.00", "0.99", "0.98"]) + os.linesep
    )
    assert (
        df2str(pd.DataFrame(data=[1, 1, 1]), digits=2, monotonicity={0: {"sign": -1}})
        == os.linesep.join(["1.00", "0.99", "0.98"]) + os.linesep
    )
    assert (
        df2str(pd.DataFrame(data=[1, 1, 1]), digits=2, monotonicity={0: {"sign": 1}})
        == os.linesep.join(["1.00", "1.01", "1.02"]) + os.linesep
    )
    assert (
        df2str(pd.DataFrame(data=[1, 1, 1]), digits=2, monotonicity={0: {"sign": 1}})
        == os.linesep.join(["1.00", "1.01", "1.02"]) + os.linesep
    )
    assert (
        df2str(pd.DataFrame(data=[1, 1, 1]), digits=7, monotonicity={0: {"sign": -1}})
        == os.linesep.join(["1.0000000", "0.9999999", "0.9999998"]) + os.linesep
    )

    # For strict monotonicity we will introduce negativity:
    dframe = pd.DataFrame(data=[0.00001, 0.0, 0.0, 0.0], columns=["PC"])
    assert (
        df2str(dframe, monotonicity={"PC": {"sign": -1}})
        == os.linesep.join(["0.0000100", "0.0000000", "-0.0000001", "-0.0000002"])
        + os.linesep
    )

    # Actual data that has occured:
    dframe = pd.DataFrame(
        data=[0.0000027, 0.0000026, 0.0000024, 0.0000024, 0.0000017], columns=["PC"]
    )
    assert (
        df2str(dframe, monotonicity={"PC": {"sign": -1}})
        == os.linesep.join(
            ["0.0000027", "0.0000026", "0.0000024", "0.0000023", "0.0000017"]
        )
        + os.linesep
    )


@pytest.mark.parametrize(
    "series, monotonicity, expected",
    [
        (
            [0.00, 0.0002, 0.01, 0.010001, 0.0100001, 0.01, 0.99, 1.0001, 1.00],
            {0: {"sign": 1, "lower": 0, "upper": 1}},
            ["0.00", "0.00", "0.01", "0.02", "0.03", "0.04", "0.99", "1.00", "1.00"],
        ),
        (
            [0.02, 0.01, 0.01, 0.00, 0.00],
            {0: {"sign": -1, "lower": 0}},
            ["0.02", "0.01", "0.00", "0.00", "0.00"],
        ),
        (
            [0.02, 0.01, 0.01, 0.00, 0.00],
            {0: {"sign": -1}},
            ["0.02", "0.01", "-0.00", "-0.01", "-0.02"],
            #                 ^ this sign is optional, allowed when negative is allowed
        ),
        ([1.0, 1.0, 1.0], {0: {"sign": 1, "upper": 1}}, ["1.00", "1.00", "1.00"]),
        ([1, 1, 1], {0: {"sign": 1}}, ["1.00", "1.01", "1.02"]),
        ([1, 1, 1], {0: {"sign": -1, "upper": 1}}, ["1.00", "1.00", "1.00"]),
        ([1, 1, 1], {0: {"sign": -1}}, ["1.00", "0.99", "0.98"]),
        ([1, 1, 1], {0: {"sign": -1, "lower": 1}}, ["1.00", "1.00", "1.00"]),
        ([0, 0, 0], {0: {"sign": -1, "allowzero": True}}, ["0.00", "0.00", "0.00"]),
        ([0, 0, 0], {0: {"sign": 1, "allowzero": True}}, ["0.00", "0.00", "0.00"]),
        (
            [1, 1, 0.5, 0.01, 0.01, 0, 0],
            {0: {"sign": -1, "lower": 0, "upper": 1}},
            ["1.00", "1.00", "0.50", "0.01", "0.00", "0.00", "0.00"],
        ),
        (
            [0, 0, 0.01, 0.01, 0.5, 1, 1],
            {0: {"sign": 1, "lower": 0, "upper": 1}},
            ["0.00", "0.00", "0.01", "0.02", "0.50", "1.00", "1.00"],
        ),
        (
            [1, 0.01, 1e-3, 0],
            {0: {"sign": -1, "lower": 0, "upper": 1}},
            ["1.00", "0.01", "0.00", "0.00"],
        ),
        (
            [1, 0.01, 1e-9, 1e-10, 0],  # Tricky due to epsilon
            {0: {"sign": -1, "lower": 0, "upper": 1}},
            ["1.00", "0.01", "0.00", "0.00", "0.00"],
        ),
        (
            [1, 0.01, 4.66e-8, 1.56e-8, 4.09e-9, 1e-10, 0],  # Tricky due to epsilon
            {0: {"sign": -1, "lower": 0, "upper": 1}},
            ["1.00", "0.01", "0.00", "0.00", "0.00", "0.00", "0.00"],
        ),
        (
            [1, 0.01, 4.78e-8, 2.09e-8, 4.09e-9, 1e-10, 0],  # Tricky due to epsilon
            {0: {"sign": -1, "lower": 0, "upper": 1}},
            ["1.00", "0.01", "0.00", "0.00", "0.00", "0.00", "0.00"],
        ),
        # Test two columns at the same time:
        (
            [[0, 0], [0, 0], [0, 0]],
            {0: {"sign": -1, "allowzero": True}, 1: {"sign": -1, "allowzero": True}},
            ["0.00 0.00", "0.00 0.00", "0.00 0.00"],
        ),
        (
            [[1, 1], [1, 1], [1, 1]],
            {0: {"sign": -1}, 1: {"sign": 1}},
            ["1.00 1.00", "0.99 1.01", "0.98 1.02"],
        ),
        # Example in docstring for modify_dframe_monotonicity()
        (
            [0.00, 0.0002, 0.01, 0.010001, 0.0100001, 0.01, 0.99, 0.999, 1.0001, 1.00],
            {0: {"sign": 1, "lower": 0, "upper": 1}},
            [
                "0.00",
                "0.00",
                "0.01",
                "0.02",
                "0.03",
                "0.04",
                "0.99",
                "1.00",
                "1.00",
                "1.00",
            ],
        ),
    ],
)
def test_df2str_nonstrict_monotonicity(series, monotonicity, expected):
    """Test that we can have non-strict monotonicity at upper and/or lower limits"""
    assert (
        df2str(
            pd.DataFrame(data=series),
            digits=2,
            monotonicity=monotonicity,
        ).splitlines()
        == expected
    )


# Test similarly for digits=1:
@pytest.mark.parametrize(
    "series, monotonicity, expected",
    [
        (
            [0.00, 0.0002, 0.01, 0.010001, 0.0100001, 0.01, 0.99, 1.0001, 1.00],
            {0: {"sign": 1, "lower": 0, "upper": 1}},
            ["0.0", "0.0", "0.0", "0.0", "0.0", "0.0", "1.0", "1.0", "1.0"],
        ),
        (
            [0.2, 0.1, 0.1, 0.0, 0.0],
            {0: {"sign": -1, "lower": 0}},
            ["0.2", "0.1", "0.0", "0.0", "0.0"],
        ),
        (
            [0.2, 0.1, 0.1, 0.0, 0.0],
            {0: {"sign": -1}},
            ["0.2", "0.1", "-0.0", "-0.1", "-0.2"],
            #               ^ this sign is optional, allowed when negative is allowed
        ),
        ([1.0, 1.0, 1.0], {0: {"sign": 1, "upper": 1}}, ["1.0", "1.0", "1.0"]),
        ([1, 1, 1], {0: {"sign": 1}}, ["1.0", "1.1", "1.2"]),
        ([1, 1, 1], {0: {"sign": -1, "upper": 1}}, ["1.0", "1.0", "1.0"]),
        ([1, 1, 1], {0: {"sign": -1}}, ["1.0", "0.9", "0.8"]),
        ([1, 1, 1], {0: {"sign": -1, "lower": 1}}, ["1.0", "1.0", "1.0"]),
        ([0, 0, 0], {0: {"sign": -1, "allowzero": True}}, ["0.0", "0.0", "0.0"]),
        ([0, 0, 0], {0: {"sign": 1, "allowzero": True}}, ["0.0", "0.0", "0.0"]),
        (
            [1, 1, 0.5, 0.01, 0.01, 0, 0],
            {0: {"sign": -1, "lower": 0, "upper": 1}},
            ["1.0", "1.0", "0.5", "0.0", "0.0", "0.0", "0.0"],
        ),
        (
            [0, 0, 0.1, 0.1, 0.5, 1, 1],
            {0: {"sign": 1, "lower": 0, "upper": 1}},
            ["0.0", "0.0", "0.1", "0.2", "0.5", "1.0", "1.0"],
        ),
        (
            [1, 0.1, 1e-2, 0],
            {0: {"sign": -1, "lower": 0, "upper": 1}},
            ["1.0", "0.1", "0.0", "0.0"],
        ),
        # Example in docstring for modify_dframe_monotonicity()
        (
            [0.00, 0.0002, 0.01, 0.010001, 0.0100001, 0.01, 0.99, 0.999, 1.0001, 1.00],
            {0: {"sign": 1, "lower": 0, "upper": 1}},
            [
                "0.0",
                "0.0",
                "0.0",
                "0.0",
                "0.0",
                "0.0",
                "1.0",
                "1.0",
                "1.0",
                "1.0",
            ],
        ),
    ],
)
def test_df2str_nonstrict_monotonicity_digits1(series, monotonicity, expected):
    """Test that we can have non-strict monotonicity at upper and/or lower limits"""
    assert (
        df2str(
            pd.DataFrame(data=series),
            digits=1,
            monotonicity=monotonicity,
        ).splitlines()
        == expected
    )


@pytest.mark.parametrize(
    "series, monotonicity",
    [
        (
            [0, 1],
            {0: {"sign": -1}},
        ),
        ([0, 0.5, 1], {0: {"sign": -1}}),
        (
            [0, 1],
            {0: {"sign": -1, "lower": 0, "upper": 1}},
        ),
        ([0, 0.5, 1], {0: {"sign": -1, "lower": 0, "upper": 1}}),
        ([0, 1], {1: "foo"}),  # not a dict of dict
        ([0, 1], {0: {}}),  # sign is required
        ([0, 1], {0: {"SGN": 1}}),  # sign is required
        ([0, 1], {0: {"sign": 2}}),  # sign have abs = 1
        ([0, 1], {0: {"sign": 1, "lower": "a"}}),
        ([0, 1], {0: {"sign": 1, "lower": 2}}),
        ([0, 1], {0: {"sign": 1, "upper": -2}}),
        ([0], {0: {"sign": -1, "upper": -1}}),
        ([0], {0: {"sign": -1, "lower": 1}}),
        ([0], {0: {"sign": 1, "upper": -1}}),
        ([0], {0: {"sign": 1, "lower": 1}}),
    ],
)
def test_df2str_nonstrict_monotonicity_valueerror(series, monotonicity):
    """Check we get ValueError in the correct circumstances"""
    with pytest.raises(ValueError):
        df2str(
            pd.DataFrame(data=series),
            digits=2,
            monotonicity=monotonicity,
        )


@pytest.mark.parametrize(
    "series, monotonicity, expected_series",
    [
        ([], {"sign": 1}, []),
        ([0], {"sign": 1}, [0]),
        ([0, 0], {"sign": 1}, [0, 0]),
        ([0, -0.1], {"sign": 1}, [0, 0]),
        ([0, -0.1, -0.1], {"sign": 1}, [0, 0, 0]),
        ([0, -0.1], {"sign": -1}, [0, -0.1]),
        ([0, -0.1, -0.1], {"sign": -1}, [0, -0.1, -0.1]),
        (np.array([0, 2]), {"sign": 1, "upper": 1}, [0, 1]),
        ([0, 2], {"sign": 1, "upper": 1, "lower": 0.1}, [0.1, 1]),
        ([0, 2], {"sign": -1, "upper": 1, "lower": 0.1}, [0.1, 0.1]),
    ],
)
def test_clip_accumulate(series, monotonicity, expected_series):
    """Test that we are able to clip to upper and lower limits, and
    use numpy's accumulate to ensure non-strict monotonicity"""
    assert (clip_accumulate(series, monotonicity) == expected_series).all()


@pytest.mark.parametrize(
    "series, monotonicity, colname, error_str",
    [
        ([], {}, "", None),
        ([], {}, "foo", None),
        ([2], {"upper": 1}, "foobar", "larger than upper limit in column foobar"),
        ([2], {"upper": 1}, "", "larger than upper limit in column"),
        ([2], {"upper": 1}, None, "larger than upper limit in column None"),
        ([2], {"lower": 3}, "foobar", "smaller than lower limit in column foobar"),
    ],
)
def test_check_limits(series, monotonicity, colname, error_str):
    """Test that we can check upper and lower limits in series
    with proper error messages."""
    if error_str is not None:
        with pytest.raises(ValueError) as err:
            check_limits(series, monotonicity, colname)
        assert error_str in str(err)
    else:
        check_limits(series, monotonicity, colname)


@pytest.mark.parametrize(
    "series, monotonicity, digits, expected",
    [
        ([], {"sign": 1}, 0, []),
        ([], {"sign": 1}, 2, []),
        ([0], {"sign": 1}, 2, [False]),
        ([0, 0.1], {"sign": 1}, 2, [False, False]),
        ([0, 0.1], {"sign": 1}, 1, [False, False]),
        ([0, 0.1], {"sign": 1}, 0, [False, True]),
        ([0.1, 0], {"sign": -1}, 2, [False, False]),
        ([0.1, 0], {"sign": -2}, 1, [False, False]),
        ([0.1, 0], {"sign": -1}, 0, [False, True]),
        # But this is allowed if at limits:
        ([0, 0.1], {"sign": 2, "upper": 0.1}, 0, [False, False]),
        ([0.1, 0], {"sign": -1, "lower": 0}, 0, [False, False]),
    ],
)
def test_rows_to_be_fixed(series, monotonicity, digits, expected):
    """Check that we can make a boolean array of which elements must be fixed for
    monotonicity"""
    assert (rows_to_be_fixed(series, monotonicity, digits) == expected).all()


@pytest.mark.parametrize(
    "series, digits, sign, expected_error",
    [
        ([], 0, 1, None),
        ([0], 0, 1, None),
        ([0, -1], 0, 1, None),
        ([0, -1], 1, 1, None),  # Assumed constant with digits=1
        ([0, -1.001], 1, 1, ValueError),  # Just not constant enough
        ([0, -1], 2, 1, ValueError),
        ([0, 1], 2, -1, ValueError),
        ([0, 1], 1, -1, None),
    ],
)
def test_check_almost_monotone(series, digits, sign, expected_error):
    """Test that we have a way to determine if a series is almost monotone, that
    is with respect to how many digits is to be looked at."""
    if expected_error is not None:
        with pytest.raises(ValueError):
            check_almost_monotone(series, digits, sign)
    else:
        check_almost_monotone(series, digits, sign)


@pytest.mark.parametrize(
    "monotonicity, dframe_colnames, error_str",
    [
        (None, [], None),
        ({}, [], None),
        ("sign", [], "monotonicity argument must be a dict"),
        ({"sign": 1}, [], "monotonicity argument must be a dict of dicts"),
        ({"foo": {"sign": 1}}, [], "Column foo does not exist in dataframe"),
        ({"foo": {"sign": 1}}, ["foo"], None),
        ({"foo": {"sgn": 1}}, ["foo"], "Unknown keys in monotonicity dict"),
        ({"foo": {"upper": 1}}, ["foo"], "Monotonocity sign not specified for foo"),
        ({"foo": {"sign": 1, "upper": 1}}, ["foo"], None),
        ({"foo": {"sign": 1, "lower": 1}}, ["foo"], None),
        ({"foo": {"sign": 1, "allowzero": True}}, ["foo"], None),
        (
            {"foo": {"sign": 1, "allowzero": "yes"}},
            ["foo"],
            "allowzero in monotonicity argument must be True/False",
        ),
        (
            {"foo": {"sign": "positive"}},
            ["foo"],
            "Monotonocity sign positive not valid",
        ),
        ({"foo": {"sign": 2}}, ["foo"], "Monotonocity sign must be -1 or +1"),
        ({"foo": {"sign": -2}}, ["foo"], "Monotonocity sign must be -1 or +1"),
    ],
)
def test_validate_monotonicity_arg(monotonicity, dframe_colnames, error_str):
    """Check error messages for monotonicity dictionaries"""
    if error_str is not None:
        with pytest.raises(ValueError) as err:
            validate_monotonicity_arg(monotonicity, dframe_colnames)
        assert error_str in str(err)
    else:
        validate_monotonicity_arg(monotonicity, dframe_colnames)


def test_modify_dframe_monotonicity_many_iterations():
    """This dataframe should be the worst kind scenario (?) in terms of how many
    iterations are needed. This gives len(dframe) - 1 iterations, and the code
    would bail (AssertionError) if we get to 2*len(dframe) iterations."""
    modify_dframe_monotonicity(
        pd.DataFrame({"SW": [0.1, 0.1, 0.1, 0.09]}), {"SW": {"sign": -1}}, 2
    )



"""Object to represent GasWater, implemented as a Container
object for one WaterOil and one GasOil object"""

from typing import Optional

import pandas as pd

from pyscal import getLogger_pyscal
from pyscal.utils.relperm import crosspoint

from .gasoil import GasOil
from .wateroil import WaterOil

logger = getLogger_pyscal(__name__)


def is_documented_by(original):
    """Decorator to avoid duplicating function docstrings"""

    def wrapper(target):
        target.__doc__ = original.__doc__
        return target

    return wrapper


class GasWater:
    """A representation of two-phase properties for gas-water

    Internally, this class handles gas-water by using one WaterOil
    object and one GasOil object, with dummy parameters for oil.

    Args:
        swirr: Irreducible water saturation for capillary pressure
        swl: First water saturation point in outputted tables.
        sgl: Minimum gas saturation in a gas paleozone.
        swcr: Critical water saturation, water is immobile below this
        sgrw: Residual gas saturation after water flooding.
        sgcr: Critical gas saturation, gas is immobile below this
        h: Saturation intervals in generated tables.
        tag: Optional text that will be included as comments.
        fast: Set to True if you prefer speed over robustness. Not recommended,
            pyscal will not guarantee valid output in this mode.
    """

    def __init__(
        self,
        swirr: float = 0.0,
        swl: float = 0.0,
        sgl: float = 0.0,
        swcr: float = 0.0,
        sgrw: float = 0.0,
        sgcr: float = 0.0,
        h: Optional[float] = None,
        tag: str = "",
        fast: bool = False,
    ) -> None:
        """Sets up the saturation range for a GasWater object,
        by initializing one WaterOil and one GasOil object, with
        endpoints set to fit with the GasWater proxy object."""
        self.fast: bool = fast

        if h is None:
            h = 0.01

        self.wateroil: WaterOil = WaterOil(
            swirr=swirr,
            swl=swl,
            swcr=swcr,
            sorw=sgrw,
            h=h,
            tag=tag,
            fast=fast,
            _sgcr=sgcr,
            _sgl=sgl,
        )

        self.sgcr: float = sgcr
        self.sgrw: float = sgrw
        self.sgl: float = sgl
        # (remaining parameters are implemented as property functions)

        self.gasoil: GasOil = GasOil(
            swirr=swirr,  # Reserved for use in capillary pressure
            sgcr=sgcr,
            sorg=0,  # Irrelevant for GasWater
            swl=swl,
            h=h,
            tag=tag,
            fast=fast,
            _sgl=sgl,
        )

        # Dummy oil curves, just to avoid objects being invalid.
        self.wateroil.add_corey_oil()
        self.gasoil.add_corey_oil()

    def selfcheck(self) -> bool:
        """Run selfcheck on the data.

        Performs tests if necessary data is ready in the object for
        printing Eclipse include files, and checks some numerical
        properties (direction and monotonicity)
        """
        if self.wateroil is not None and self.gasoil is not None:
            return self.wateroil.selfcheck() and self.gasoil.selfcheck()
        logger.error("None objects in GasWater (bug)")
        return False

    def add_corey_water(
        self, nw: float = 2.0, krwend: float = 1.0, krwmax: Optional[float] = None
    ) -> None:
        """Add krw data through the Corey parametrization

        A column named 'krw' will be added. If it exists, it will
        be replaced.

        The Corey model applies for sw < 1 - sgrw. For higher
        water saturations, krw is linear between krwend and krwmax.

        krwmax will be ignored if sgrw is close to zero

        Args:
            nw: Corey parameter for water.
            krwend: value of krw at 1 - sgcr.
            krwmax: maximal value at Sw=1. Default 1
        """
        self.wateroil.add_corey_water(nw, krwend, krwmax)

    def add_corey_gas(self, ng: float = 2.0, krgend: float = 1.0) -> None:
        """Add krg data through the Corey parametrization

        A column named 'krg' will be added. If it exists, it will
        be replaced.

        Args:
            ng: Corey parameter for gas
            krgend: value of krg at swl.
        """
        self.gasoil.add_corey_gas(ng, krgend, krgmax=None)

    def add_LET_water(
        self,
        l: float = 2.0,
        e: float = 2.0,
        t: float = 2.0,
        krwend: float = 1.0,
        krwmax: Optional[float] = None,
    ) -> None:
        """Add krw data through LET parametrization

        The LET model applies for sw < 1 - sgrw. For higher
        water saturations, krw is linear between krwend and krwmax.

        krwmax will be ignored if sorw is close to zero.

        Args:
            l: LET parameter
            e: LET parameter
            t: LET parameter
            krwend: value of krw at 1 - sorw
            krwmax: maximal value at Sw=1. Default 1
        """
        self.wateroil.add_LET_water(l, e, t, krwend, krwmax)

    def add_LET_gas(
        self, l: float = 2.0, e: float = 2.0, t: float = 2.0, krgend: float = 1.0
    ) -> None:
        """Add krg data through the LET parametrization

        A column named 'krg' will be added. If it exists, it will
        be replaced.

        Args:
            l: LET parameter
            e: LET parameter
            t: LET parameter
            krgend: value of krg at swl
        """
        self.gasoil.add_LET_gas(l, e, t, krgend, krgmax=None)

    @is_documented_by(WaterOil.add_simple_J)
    def add_simple_J(
        self,
        a: float = 5.0,
        b: float = -1.5,
        poro_ref: float = 0.25,
        perm_ref: float = 100.0,
        drho: float = 300.0,
        g: float = 9.81,
    ) -> None:
        """Add a simple J-function, handed over to the WaterOil object"""
        self.wateroil.add_simple_J(a, b, poro_ref, perm_ref, drho, g)

    @is_documented_by(WaterOil.add_simple_J_petro)
    def add_simple_J_petro(
        self,
        a: float,
        b: float,
        poro_ref: float = 0.25,
        perm_ref: float = 100.0,
        drho: float = 300.0,
        g: float = 9.81,
    ) -> None:
        """Add a simple petrophysical variant of the J-function, handed
        over to the WaterOil object"""
        self.wateroil.add_simple_J_petro(a, b, poro_ref, perm_ref, drho, g)

    def SWFN(self, header: bool = True, dataincommentrow: bool = True):
        """Produce SWFN input to Eclipse

        The columns sw, krw and pc are outputted and formatted
        accordingly.

        Meta-information for the tabulated data are printed
        as Eclipse comments.

        Args:
            header: boolean for whether the SWFN string should be emitted.
                If you have multiple satnums, you should have True only
                for the first (or False for all, and emit the SWFN yourself).
                Defaults to True.
            dataincommentrow: boolean for wheter metadata should be printed,
                defaults to True.
        """
        if self.fast:
            crosspointcomment = ""
        else:
            crosspoint_value = self.crosspoint()
            if crosspoint_value is not None:
                crosspointcomment = f"-- krw = krg @ sw={crosspoint_value:1.5f}\n"
            else:
                crosspointcomment = ""
        return self.wateroil.SWFN(
            header,
            dataincommentrow,
            swcomment=self.swcomment,
            crosspointcomment=crosspointcomment,
        )

    def SGFN(self, header: bool = True, dataincommentrow: bool = True):
        """Produce SGFN input for Eclipse reservoir simulator.

        The columns sg and krg are outputted and formatted accordingly.

        Meta-information for the tabulated data are printed as Eclipse comments.

        Args:
            header: boolean for whether the SGFN string should be emitted.
                If you have multiple satnums, you should have True only
                for the first (or False for all, and emit the SGFN yourself).
                Defaults to True.
            dataincommentrow: boolean for wheter metadata should be printed,
                defaults to True.
        """
        if self.fast:
            crosspointcomment = ""
        else:
            crosspoint_value = self.crosspoint()
            if crosspoint_value is not None:
                crosspointcomment = f"-- krw = krg @ sw={crosspoint_value:1.5f}\n"
            else:
                crosspointcomment = ""
        return self.gasoil.SGFN(
            header,
            dataincommentrow,
            sgcomment=self.sgcomment,
            crosspointcomment=crosspointcomment,
        )

    def crosspoint(self) -> Optional[float]:
        """Calculate the sw value where krg == krw.

        Accuracy of this crosspoint depends on the resolution chosen
        when initializing the saturation range (it uses linear
        interpolation to solve for the zero)

        Returns:
            The gas saturation where krw == krg, for relperm linearly
            interpolated in water saturation.
        """
        if not {"SW", "KRW"}.issubset(self.wateroil.table.columns):
            logger.warning("Can't compute crosspoint when KRW is not present")
            return None
        if not {"SL", "KRG"}.issubset(self.gasoil.table.columns):
            logger.warning("Can't compute crosspoint when KRG is not present")
            return None
        dframe = pd.concat(
            [self.wateroil.table[["SW", "KRW"]], self.gasoil.table[["SL", "KRG"]]],
            sort=False,
        )
        # The  "SL" column in the GasOil object corresponds exactly to "SW" in WaterOil
        # but since they are floating point, we do not want to "merge" dataframes on it,
        # rather concatenate and let linear interpolation fill in values.
        dframe["SW"] = dframe["SW"].fillna(value=0)
        dframe["SL"] = dframe["SL"].fillna(value=0)
        dframe["sat"] = dframe["SL"] + dframe["SW"]
        dframe = (
            dframe.set_index("sat")
            .sort_index()
            .interpolate(method="slinear")
            .dropna()
            .reset_index()
        )
        return crosspoint(dframe, "sat", "KRW", "KRG")

    def plotkrwkrg(
        self,
        mpl_ax=None,
        color: str = "blue",
        alpha: float = 1,
        linewidth: int = 1,
        linestyle: str = "-",
        marker: Optional[str] = None,
        label: str = "",
        logyscale: bool = False,
    ):
        """Plot krw and krg

        If the argument 'mpl_ax' is not supplied, a new plot
        window will be made. If supplied, it will draw on
        the specified axis."""

        # Lazy import of matplotlib for speed reasons.
        import matplotlib  # noqa: PLC0415
        import matplotlib.pyplot as plt  # noqa: PLC0415

        if mpl_ax is None:
            matplotlib.style.use("ggplot")
            _, useax = plt.subplots()
        else:
            useax = mpl_ax
        if logyscale:
            useax.set_yscale("log")
            useax.set_ylim((1e-8, 1))
        self.wateroil.table.plot(
            ax=useax,
            x="SW",
            y="KRW",
            c=color,
            alpha=alpha,
            legend=None,
            label=label,
            linewidth=linewidth,
            linestyle=linestyle,
            marker=marker,
        )
        self.gasoil.table.plot(
            ax=useax,
            x="SL",
            y="KRG",
            c=color,
            alpha=alpha,
            label=None,
            legend=None,
            linewidth=linewidth,
            linestyle=linestyle,
            marker=marker,
        )
        plt.xlabel("SW")
        if mpl_ax is None:
            plt.show()

    @property
    def swirr(self) -> float:
        """Get the swirr, irreducible water saturation used for pc-init"""
        return self.wateroil.swirr

    @property
    def swl(self) -> float:
        """Get the swl"""
        return self.wateroil.swl

    @property
    def swcr(self) -> float:
        """Get the swcr"""
        return self.wateroil.swcr

    @property
    def swcomment(self) -> str:
        """Get a string representation of the endpoints used for water"""
        return self.wateroil.swcomment.replace("sorw", "sgrw")

    @property
    def sgcomment(self) -> str:
        """Get a string representation of the endpoints used for gas"""
        return (
            self.gasoil.sgcomment.replace("sorg", "sgrw")
            .replace(", krgendanchor=sgrw", "")
            .replace(", krgendanchor=", "")
        )

    @property
    def krwcomment(self) -> str:
        """Get a string representation describing krw"""
        return self.wateroil.krwcomment

    @property
    def krgcomment(self) -> str:
        """Get a string representation describing krg"""
        return self.gasoil.krgcomment

    @property
    def tag(self) -> str:
        """Get the user configured tag"""
        if self.wateroil.tag == self.gasoil.tag:
            return self.wateroil.tag
        raise ValueError("Internal tag-inconsistency in GasWater")
